import requests

from . import services


class PlatformException(Exception):
    """
    Dataloop Exceptions
    """
    def __init__(self, error=None, message=None):
        """
        :param error: 'str' status code  out of list
        :param message: `str` describing the error
        """
        self.status_code = None
        self.message = message

        exceptions = {
            '400': BadRequest,
            '401': Unauthorized,
            '403': Forbidden,
            '404': NotFound,
            '408': RequestTimeout,
            '409': Conflict,
            '424': FailedDependency,
            '500': InternalServerError,
            '600': TokenExpired,
            '1001': ShowAnnotationError,
            '1002': ExportAnnotationError,
            '2001': MissingEntity,
            '3001': SDKError,
            '3002': EntityPrintError
        }

        if isinstance(error, requests.models.Response) or isinstance(error, services.AsyncResponse):
            if hasattr(error, 'status_code'):
                self.status_code = str(error.status_code)
            if hasattr(error, 'text'):
                try:
                    self.message = error.json().get('message', error.text)
                except Exception:
                    self.message = error.text

        else:
            self.status_code = error
            self.message = message

        if self.status_code in exceptions:
            raise exceptions[self.status_code](status_code=self.status_code, message=self.message)
        else:
            raise UnknownException(status_code=self.status_code, message=self.message)


class ExceptionMain(Exception):
    def __init__(self, status_code='Unknown Status Code', message='Unknown Error Message'):
        self.status_code = status_code
        self.message = message
        super().__init__(status_code, message)


class NotFound(ExceptionMain):
    pass


class InternalServerError(ExceptionMain):
    pass


class Forbidden(ExceptionMain):
    pass


class Conflict(ExceptionMain):
    pass


class FailedDependency(ExceptionMain):
    pass


class BadRequest(ExceptionMain):
    pass


class Unauthorized(ExceptionMain):
    pass


class RequestTimeout(ExceptionMain):
    pass


class TokenExpired(ExceptionMain):
    pass


class UnknownException(ExceptionMain):
    pass


class MissingEntity(ExceptionMain):
    pass


##########################
# annotations exceptions #
##########################
class ShowAnnotationError(ExceptionMain):
    """ raised when error in annotations drawing"""
    pass


class ExportAnnotationError(ExceptionMain):
    """ raised when error in annotations drawing"""
    pass


class SDKError(ExceptionMain):
    """ raised when error in annotations drawing"""
    pass


class EntityPrintError(ExceptionMain):
    """ raised when error in annotations drawing"""
    pass



from collections import namedtuple
import traceback
import logging
import random
import uuid
import attr
import os

from .. import entities, PlatformException, repositories, exceptions
from ..services.api_client import ApiClient
from .label import Label

logger = logging.getLogger(name='dtlpy')


class AttributesTypes:
    CHECKBOX = "checkbox"
    RADIO_BUTTON = "radio_button"
    SLIDER = "range"
    YES_NO = "boolean"
    FREE_TEXT = "freeText"

class AttributesRange:
    def __init__(self, min_range, max_range, step):
        self.min_range = min_range
        self.max_range = max_range
        self.step = step

    def to_json(self):
        return {'min': self.min_range, 'max': self.max_range, 'step': self.step}


class LabelHandlerMode:
    ADD = "add"
    UPDATE = "update"
    UPSERT = "upsert"


@attr.s
class Ontology(entities.BaseEntity):
    """
    Ontology object
    """
    # api
    _client_api = attr.ib(type=ApiClient, repr=False)

    # params
    id = attr.ib()
    creator = attr.ib()
    url = attr.ib(repr=False)
    title = attr.ib()
    labels = attr.ib(repr=False)
    metadata = attr.ib(repr=False)
    attributes = attr.ib()

    # entities
    _recipe = attr.ib(repr=False, default=None)
    _dataset = attr.ib(repr=False, default=None)
    _project = attr.ib(repr=False, default=None)

    # repositories
    _repositories = attr.ib(repr=False)

    # defaults
    _instance_map = attr.ib(default=None, repr=False)
    _color_map = attr.ib(default=None, repr=False)

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['ontologies', 'datasets', 'projects'])

        if self._recipe is None:
            ontologies = repositories.Ontologies(client_api=self._client_api, recipe=self._recipe)
        else:
            ontologies = self.recipe.ontologies

        r = reps(ontologies=ontologies, datasets=repositories.Datasets(client_api=self._client_api),
                 projects=repositories.Projects(client_api=self._client_api))
        return r

    @property
    def recipe(self):
        if self._recipe is None:
            filters = entities.Filters(resource=entities.FiltersResource.RECIPE)
            filters.add(field="ontologies", values=self.id)
            recipes = self.project.recipes.list(filters=filters)
            if recipes.items_count > 0:
                self._recipe = recipes.items[0]
            else:
                logger.warning(f"Ontology ID: {self.id} Does not belong to a recipe")
        if self._recipe is not None:
            assert isinstance(self._recipe, entities.Recipe)
        return self._recipe

    @property
    def dataset(self):
        if self._dataset is None:
            if self.recipe is not None:
                self._dataset = self.recipe.dataset
        if self._dataset is not None:
            assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def project(self):
        if self._project is None:
            if 'system' in self.metadata:
                project_id = self.metadata['system'].get('projectIds', None)
                if project_id is not None:
                    self._project = self.projects.get(project_id=project_id[0])
            elif self.dataset is not None:
                self._project = self.dataset.project
        if self._project is not None:
            assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def ontologies(self):
        if self._repositories.ontologies is not None:
            assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    @property
    def projects(self):
        if self._repositories.projects is not None:
            assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def labels_flat_dict(self):
        flatten_dict = dict()

        def add_to_dict(tag, father):
            flatten_dict[tag] = father
            for child in father.children:
                add_to_dict('{}.{}'.format(tag, child.tag), child)

        for label in self.labels:
            add_to_dict(label.tag, label)
        return flatten_dict

    @property
    def instance_map(self):
        """
         instance mapping for creating instance mask

        :return dictionary {label: map_id}
        :rtype: dict
        """
        if self._instance_map is None:
            labels = [label for label in self.labels_flat_dict]
            labels.sort()
            # each label gets index as instance id
            self._instance_map = {label: (i_label + 1) for i_label, label in enumerate(labels)}
        return self._instance_map

    @instance_map.setter
    def instance_map(self, value: dict):
        """
        instance mapping for creating instance mask

        :param value: dictionary {label: map_id}
        :rtype: dict
        """
        if not isinstance(value, dict):
            raise ValueError('input must be a dictionary of {label_name: instance_id}')
        self._instance_map = value

    @property
    def color_map(self):
        """
        Color mapping of labels, {label: rgb}

        :return: dict
        :rtype: dict
        """
        if self._color_map is None:
            self._color_map = {k: v.rgb for k, v in self.labels_flat_dict.items()}
        return self._color_map

    @color_map.setter
    def color_map(self, values):
        """
        Color mapping of labels, {label: rgb}

        :param values: dict {label: rgb}
        :return:
        """
        if not isinstance(values, dict):
            raise ValueError('input must be a dict. got: {}'.format(type(values)))
        self._color_map = values

    @staticmethod
    def _protected_from_json(_json, client_api, recipe=None, dataset=None, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            ontology = Ontology.from_json(_json=_json,
                                          client_api=client_api,
                                          project=project,
                                          dataset=dataset,
                                          recipe=recipe,
                                          is_fetched=is_fetched)
            status = True
        except Exception:
            ontology = traceback.format_exc()
            status = False
        return status, ontology

    @property
    def _use_attributes_2(self):
        if isinstance(self.metadata, dict):
            attributes = self.metadata.get("attributes", None)
            if attributes is not None:
                return True
            else:
                if isinstance(self.attributes, list) and len(self.attributes) > 0:
                    return False
        return True

    @classmethod
    def from_json(cls, _json, client_api, recipe=None, dataset=None, project=None, is_fetched=True):
        """
        Build an Ontology entity object from a json

        :param bool is_fetched: is Entity fetched from Platform
        :param dtlpy.entities.project.Project project: project entity
        :param dtlpy.entities.dataset.Dataset dataset: dataset
        :param dict _json: _json response from host
        :param dtlpy.entities.recipe.Recipe recipe: ontology's recipe
        :param dl.ApiClient client_api: ApiClient entity
        :return: Ontology object
        :rtype: dtlpy.entities.ontology.Ontology
        """
        attributes_v2 = _json.get('metadata', {}).get("attributes", [])
        attributes_v1 = _json.get("attributes", [])
        attributes = attributes_v2 if attributes_v2 else attributes_v1

        labels = list()
        for root in _json["roots"]:
            labels.append(entities.Label.from_root(root=root))

        inst = cls(
            metadata=_json.get("metadata", None),
            creator=_json.get("creator", None),
            url=_json.get("url", None),
            id=_json["id"],
            title=_json.get("title", None),
            attributes=attributes,
            client_api=client_api,
            project=project,
            dataset=dataset,
            recipe=recipe,
            labels=labels,
        )

        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        roots = [label.to_root() for label in self.labels]
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Ontology)._client_api,
                                                              attr.fields(Ontology)._recipe,
                                                              attr.fields(Ontology)._project,
                                                              attr.fields(Ontology)._dataset,
                                                              attr.fields(Ontology)._instance_map,
                                                              attr.fields(Ontology)._color_map,
                                                              attr.fields(Ontology)._repositories))
        _json["roots"] = roots
        return _json

    def delete(self):
        """
        Delete recipe from platform

        :return: True
        """
        return self.ontologies.delete(self.id)

    def update(self, system_metadata=False):
        """
        Update Ontology attribute

        :param bool system_metadata: bool - True, if you want to change metadata system
        :return: Ontology object
        """
        return self.ontologies.update(self, system_metadata=system_metadata)

    def _add_children(self, label_name, children, labels_node, mode):
        for child in children:
            if not isinstance(child, entities.Label):
                if isinstance(child, dict):
                    if "label_name" in child:
                        child = dict(child)
                        child["label_name"] = "{}.{}".format(label_name, child["label_name"])
                        labels_node += self._base_labels_handler(labels=[child], update_ontology=False, mode=mode)
                    else:
                        raise PlatformException("400",
                                                "Invalid parameters - child list must have label name attribute")
                else:
                    raise PlatformException("400", "Invalid parameters - child must be a dict type")
            else:
                child.tag = "{}.{}".format(label_name, child.tag)
                labels_node += self._base_labels_handler(labels=child, update_ontology=False, mode=mode)

        return labels_node

    def _labels_handler_update_mode(self, json_req, upsert=False, log_error=True):
        json_req['upsert'] = upsert
        success, response = self._client_api.gen_request(req_type="PATCH",
                                                         path="/ontologies/%s/labels" % self.id,
                                                         json_req=json_req,
                                                         log_error=log_error)
        if success:
            logger.debug("Labels {} has been added successfully".format(json_req))
        else:
            raise exceptions.PlatformException(response)
        return response

    def _labels_handler_add_mode(self, json_req):
        success, response = self._client_api.gen_request(req_type="PATCH",
                                                         path="/ontologies/%s/addLabels" % self.id,
                                                         json_req=json_req)
        if success:
            logger.debug("Labels {} has been added successfully".format(json_req))
        else:
            raise exceptions.PlatformException(response)
        return response

    def _base_labels_handler(self, labels, update_ontology=True, mode=LabelHandlerMode.UPSERT):
        """
        Add a single label to ontology using add label endpoint , nested label is also supported

        :param labels = list of labels
        :param update_ontology - return json_req if False
        :param mode add, update or upsert, relevant on update_ontology=True only
        :return: Ontology updated entire label entity
        """
        labels_node = list()
        if mode not in [LabelHandlerMode.ADD,
                        LabelHandlerMode.UPDATE,
                        LabelHandlerMode.UPSERT]:
            raise ValueError('mode must be on of: "add", "update", "upsert"')

        if not isinstance(labels, list):  # for case that add label get one label
            labels = [labels]

        for label in labels:
            if isinstance(label, str):
                # Generate label from string
                label = entities.Label(tag=label)
            elif isinstance(label, dict):
                # Generate label from dict
                label = Label.from_root(label)
            elif isinstance(label, entities.Label):
                ...
            else:
                raise ValueError(
                    'Unsupported type for `labels`. Expected a list of (str, dict, dl.Label). Got: {}'.format(
                        type(label)))

            # label entity
            label_node = {"tag": label.tag}
            if label.color is not None:
                label_node["color"] = label.hex
            if label.attributes is not None:
                label_node["attributes"] = label.attributes
            if label.display_label is not None:
                label_node["displayLabel"] = label.display_label
            if label.display_data is not None:
                label_node["displayData"] = label.display_data
            labels_node.append(label_node)
            children = label.children
            self._add_children(label.tag, children, labels_node, mode=mode)

        if not update_ontology or not len(labels_node):
            return labels_node

        json_req = {
            "labelsNode": labels_node
        }

        if mode == LabelHandlerMode.UPDATE:
            response = self._labels_handler_update_mode(json_req)
        else:
            response = self._labels_handler_update_mode(json_req, upsert=True, log_error=False)

        added_label = list()
        if "roots" not in response.json():
            raise exceptions.PlatformException("error fetching updated labels from server")

        for root in response.json()["roots"]:  # to get all labels
            added_label.append(entities.Label.from_root(root=root))

        self.labels = added_label
        return added_label

    def _add_image_label(self, icon_path):
        display_data = dict()
        if self.project is not None:
            dataset = self.project.datasets._get_binaries_dataset()
        elif self.dataset is not None:
            dataset = self.dataset.project.datasets._get_binaries_dataset()
        else:
            raise ValueError('must have project or dataset to create with icon path')
        platform_path = "/.dataloop/ontologies/{}/labelDisplayImages/".format(self.id)
        basename = os.path.basename(icon_path)
        item = dataset.items.upload(local_path=icon_path,
                                    remote_path=platform_path,
                                    remote_name='{}-{}'.format(uuid.uuid4().hex, basename))
        display_data['displayImage'] = dict()
        display_data['displayImage']['itemId'] = item.id
        display_data['displayImage']['datasetId'] = item.dataset_id
        return display_data

    def _label_handler(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                       add=True, icon_path=None, update_ontology=False, mode=LabelHandlerMode.UPSERT):
        """
        Add a single label to ontology

        :param label_name: label name
        :param color: optional - if not given a random color will be selected
        :param children: optional - children
        :param attributes: optional - attributes
        :param display_label: optional - display_label
        :param label: label
        :param add:to add or not
        :param icon_path: path to image to be display on label
        :param update_ontology: update the ontology, default = False for backward compatible
        :param mode add, update or upsert, relevant on update_ontology=True only
        :return: Label entity
        """

        if update_ontology:
            if isinstance(label, entities.Label) or isinstance(label, str):
                return self._base_labels_handler(labels=label,
                                                 update_ontology=update_ontology,
                                                 mode=mode)
            else:
                display_data = dict()
                if icon_path is not None:
                    display_data = self._add_image_label(icon_path=icon_path)
                return self._base_labels_handler(labels={"tag": label_name,
                                                         "displayLabel": display_label,
                                                         "color": color,
                                                         "attributes": attributes,
                                                         "children": children,
                                                         "displayData": display_data
                                                         },
                                                 update_ontology=update_ontology,
                                                 mode=mode)

        if not isinstance(label, entities.Label):
            if "." in label_name:
                raise PlatformException("400",
                                        "Invalid parameters - nested label can work with update_ontology option only")

            if attributes is None:
                attributes = list()
            if not isinstance(attributes, list):
                attributes = [attributes]

            # get random color if none given
            if color is None:
                color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

            if children is None:
                children = list()
            if not isinstance(children, list):
                children = [children]

            # add children    
            added_children = list()
            for child in children:
                if not isinstance(child, entities.Label):
                    added_children.append(self._label_handler(**child, add=False))
                else:
                    added_children.append(child)

            if display_label is None:
                display_label = ""
                if len(label_name.split("_")) == 1:
                    display_label = label_name[0].upper() + label_name[1:]
                else:
                    for word in label_name.split("_"):
                        display_label += word[0].upper() + word[1:] + " "
                    display_label = display_label[0:-1]

            display_data = dict()
            if icon_path is not None:
                display_data = self._add_image_label(icon_path=icon_path)

            root = {
                "value": {
                    "tag": label_name,
                    "displayLabel": display_label,
                    "color": color,
                    "attributes": attributes,
                    "displayData": display_data
                },
                "children": list(),
            }
            added_label = entities.Label.from_root(root)
            added_label.children = added_children
        else:
            added_label = label
        if add and self._validate_label(added_label=added_label, mode=mode, color=color,
                                        children=children, attributes=attributes,
                                        display_label=display_label, display_data=icon_path):
            self.labels.append(added_label)
        self._base_labels_handler(labels=added_label, update_ontology=True, mode=mode)
        return added_label

    def _validate_label(self, added_label, mode=LabelHandlerMode.UPSERT, color=None, children=None, attributes=None,
                        display_label=None, display_data=None):
        """
        check if the label is exist
        """
        for i in range(len(self.labels)):
            if self.labels[i].tag == added_label.tag:
                if mode == LabelHandlerMode.UPDATE:
                    if color:
                        self.labels[i].color = added_label.color
                    if children:
                        self.labels[i].children = added_label.children
                    if attributes:
                        self.labels[i].attributes = added_label.attributes
                    if display_label:
                        self.labels[i].display_label = added_label.display_label
                    if display_data:
                        self.labels[i].display_data = added_label.display_data
                return False
        return True

    def _labels_handler(self, label_list, update_ontology=False, mode=LabelHandlerMode.UPSERT):
        """
        Adds a list of labels to ontology

        :param list label_list: a list of labels to add to the dataset's ontology. each value should be a dict, dl.Label or a string
                        if dictionary, should look like this: {"value": {"tag": "name of the label", "displayLabel": "display name on the platform",
                                            "color": "#hex value", "attributes": [attributes]}, "children": [children]}
        :param update_ontology: update the ontology, default = False for backward compatible
        :param mode add, update or upsert, relevant on update_ontology=True only
        :return: List of label entities added
        """
        if update_ontology:
            return self._base_labels_handler(labels=label_list, mode=mode)
        labels = list()
        for label in label_list:

            if isinstance(label, str):
                label = entities.Label(tag=label)

            if isinstance(label, entities.Label):
                # label entity
                labels.append(label)
            else:
                # dictionary
                labels.append(Label.from_root(label))
        added_labels = list()
        for label in labels:
            added_labels.append(self._label_handler(label.tag, label=label, update_ontology=update_ontology))

        return added_labels

    def delete_labels(self, label_names):
        """
        Delete labels from ontology

        :param label_names: label object/ label name / list of label objects / list of label names
        :return:
        """
        if not isinstance(label_names, list):
            label_names = [label_names]

        if isinstance(label_names[0], entities.Label):
            label_names = [label.tag for label in label_names]

        for label in label_names:
            self.__delete_label(label)

        self.update()

    def __delete_label(self, label_name):
        if label_name in self.instance_map.keys():
            labels = self.labels
            label_chain = label_name.split('.')
            while len(label_chain) > 1:
                label_name = label_chain.pop(0)
                for i_label, label in enumerate(labels):
                    if label.tag == label_name:
                        labels = labels[i_label].children
                        break
            label_name = label_chain[0]
            for i_label, label in enumerate(labels):
                if label.tag == label_name:
                    labels.pop(i_label)

    def add_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                  add=True, icon_path=None, update_ontology=False):
        """
        Add a single label to ontology

        :param str label_name: str - label name
        :param tuple color: color
        :param children: children (sub labels)
        :param list attributes: attributes
        :param str display_label: display_label
        :param dtlpy.entities.label.Label label: label
        :param bool add: to add or not
        :param str icon_path: path to image to be display on label
        :param bool update_ontology: update the ontology, default = False for backward compatible
        :return: Label entity
        :rtype: dtlpy.entities.label.Label

        **Example**:

        .. code-block:: python

            label = ontology.add_label(label_name='person', color=(34, 6, 231), attributes=['big', 'small'])
        """
        return self._label_handler(label_name=label_name, color=color, children=children, attributes=attributes,
                                   display_label=display_label, label=label, add=add, icon_path=icon_path,
                                   update_ontology=update_ontology)

    def add_labels(self, label_list, update_ontology=False):
        """
        Adds a list of labels to ontology

        :param list label_list: list of labels [{"value": {"tag": "tag", "displayLabel": "displayLabel",
                                            "color": "#color", "attributes": [attributes]}, "children": [children]}]
        :param bool update_ontology: update the ontology, default = False for backward compatible
        :return: List of label entities added

        **Example**:

        .. code-block:: python

            labels = ontology.add_labels(label_list=label_list)
        """
        self._labels_handler(label_list=label_list, update_ontology=update_ontology, mode=LabelHandlerMode.UPSERT)

    def update_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                     add=True, icon_path=None, upsert=False, update_ontology=False):
        """
        Update a single label to ontology

        :param str label_name: str - label name
        :param tuple color: color
        :param children: children (sub labels)
        :param list attributes: attributes
        :param str display_label: display_label
        :param dtlpy.entities.label.Label label: label
        :param bool add: to add or not
        :param str icon_path: path to image to be display on label
        :param bool update_ontology: update the ontology, default = False for backward compatible
        :param bool upsert: if True will add in case it does not existing
        :return: Label entity
        :rtype: dtlpy.entities.label.Label

        **Example**:

        .. code-block:: python

            label = ontology.update_label(label_name='person', color=(34, 6, 231), attributes=['big', 'small'])
        """
        if upsert:
            mode = LabelHandlerMode.UPSERT
        else:
            mode = LabelHandlerMode.UPDATE

        return self._label_handler(label_name=label_name, color=color, children=children,
                                   attributes=attributes, display_label=display_label, label=label,
                                   add=add, icon_path=icon_path, update_ontology=update_ontology, mode=mode)

    def update_labels(self, label_list, upsert=False, update_ontology=False):
        """
        Update a list of labels to ontology

        :param list label_list: list of labels [{"value": {"tag": "tag", "displayLabel": "displayLabel", "color": "#color", "attributes": [attributes]}, "children": [children]}]
        :param bool upsert: if True will add in case it does not existing
        :param bool update_ontology: update the ontology, default = False for backward compatible

        :return: List of label entities added

        **Example**:

        .. code-block:: python

            labels = ontology.update_labels(label_list=label_list)
        """

        if upsert:
            mode = LabelHandlerMode.UPSERT
        else:
            mode = LabelHandlerMode.UPDATE
        self._labels_handler(label_list=label_list, update_ontology=update_ontology, mode=mode)

    def update_attributes(self,
                          title: str,
                          key: str,
                          attribute_type,
                          scope: list = None,
                          optional: bool = None,
                          values: list = None,
                          attribute_range=None):
        """
        ADD a new attribute or update if exist

        :param str title: attribute title
        :param str key: the key of the attribute must br unique
        :param AttributesTypes attribute_type: dl.AttributesTypes your attribute type
        :param list scope: list of the labels or * for all labels
        :param bool optional: optional attribute
        :param list values: list of the attribute values ( for checkbox and radio button)
        :param dict or AttributesRange attribute_range: dl.AttributesRange object
        :return: true in success
        :rtype: bool
        """
        return self.ontologies.update_attributes(
            ontology_id=self.id,
            title=title,
            key=key,
            attribute_type=attribute_type,
            scope=scope,
            optional=optional,
            values=values,
            attribute_range=attribute_range)

    def delete_attributes(self, keys: list):
        """
        Delete a bulk of attributes

        :param list keys: Keys of attributes to delete
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = ontology.delete_attributes(['1'])
        """

        return self.ontologies.delete_attributes(ontology_id=self.id, keys=keys)

    def copy_from(self, ontology_json: dict):
        """
        Import ontology to the platform.\n
        Notice: only the following fields will be updated: `labels`, `attributes`, `instance_map` and `color_map`.

        :param dict ontology_json: The source ontology json to copy from
        :return: Ontology object: The updated ontology entity
        :rtype: dtlpy.entities.ontology.Ontology

        **Example**:

        .. code-block:: python

            ontology = ontology.import_ontology(ontology_json=ontology_json)
        """
        # TODO: Add support for import from ontology entity in the Future
        if not self._use_attributes_2:
            raise ValueError("This method is only supported for attributes 2 mode!")
        new_ontology = self.from_json(_json=ontology_json, client_api=self._client_api)

        # Update 'labels' and 'attributes'
        self.labels = new_ontology.labels
        new_attributes = new_ontology.attributes
        if isinstance(new_attributes, list):
            for new_attribute in new_attributes:
                attribute_range = new_attribute.get("range", None)
                if attribute_range is not None:
                    attribute_range = entities.AttributesRange(
                        min_range=attribute_range.get("min", None),
                        max_range=attribute_range.get("max", None),
                        step=attribute_range.get("step", None)
                    )
                script_data = new_attribute.get("scriptData", None)
                if script_data is None:
                    new_attribute_key = new_attribute.get("key", None)
                    raise Exception(f"Attribute '{new_attribute_key}' scriptData is missing in the ontology json!")
                self.update_attributes(
                    title=script_data.get("title", None),
                    key=new_attribute.get("key", None),
                    attribute_type=new_attribute.get("type", None),
                    scope=new_attribute.get("scope", None),
                    optional=script_data.get("optional", None),
                    values=new_attribute.get("values", None),
                    attribute_range=attribute_range
                )

        # Get remote updated 'attributes'
        self.metadata["attributes"] = self.ontologies.get(ontology_id=self.id).attributes

        # Update 'instance map' and 'color map'
        self._instance_map = new_ontology.instance_map
        self._color_map = new_ontology.color_map
        return self.update(system_metadata=True)



from abc import ABC

from base_model_adapter import BaseModelAdapter
from .. import entities


class BaseFeatureExtractorAdapter(BaseModelAdapter, ABC):
    def __int__(self, model_entity: entities.Model = None):
        super().__init__(model_entity)

    def extract_features(self, batch: list, **kwargs):
        """ Runs inference with the model, but does not predict. Instead, extracts features for the input batch.

            Virtual method - need to implement

        :param batch: `list` a list containing a batch of items whose features will be extracted
        """
        raise NotImplementedError("Please implement 'extract_features' method in {}".format(self.__class__.__name__))

    def extract_dataset_features(self, dataset: entities.Dataset, **kwargs):
        """ Runs inference to extract features for all items in a dataset.

            Virtual method - need to implement

        :param dataset: `entities.Dataset` dataset entity whose items will have their features extracted
        """
        raise NotImplementedError("Please implement 'extract_dataset_features' method in "
                                  "{}".format(self.__class__.__name__))


import logging

from .. import exceptions, entities, miscellaneous, _api_reference, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Features:
    """
    Features repository
    """
    URL = '/features/vectors'

    def __init__(self, client_api: ApiClient,
                 project: entities.Project = None,
                 project_id: str = None,
                 item: entities.Item = None,
                 annotation: entities.Annotation = None,
                 feature_set: entities.FeatureSet = None,
                 dataset: entities.Dataset = None):
        if project is not None and project_id is None:
            project_id = project.id
        self._dataset = dataset
        self._project = project
        self._project_id = project_id
        self._item = item
        self._annotation = annotation
        self._feature_set = feature_set
        self._client_api = client_api

    ############
    # entities #
    ############
    @property
    def feature_set(self) -> entities.FeatureSet:
        return self._feature_set

    @property
    def project(self) -> entities.Project:
        if self._project is None and self._project_id is None and self._item is not None:
            self._project = self._item.project
            self._project_id = self._project.id
        if self._project is None and self._project_id is not None:
            # get from id
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)
        if self._project is None:
            # try get checkout
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Project entity in Features repository.'
                        ' Please checkout or set a project')
        assert isinstance(self._project, entities.Project)
        return self._project

    ###########
    # methods #
    ###########
    def _list(self, filters: entities.Filters):
        """
        Get dataset feature vectors list. This is a browsing endpoint, for any given path feature count will be returned,
        user is expected to perform another request then for every folder item to actually get the item list.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: json response
        """
        # prepare request
        success, response = self._client_api.gen_request(req_type="POST",
                                                         path="{}/query".format(self.URL),
                                                         json_req=filters.prepare(),
                                                         headers={'user_query': filters._user_query}
                                                         )
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/features/vectors', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List of features

        :param dtlpy.entities.filters.Filters filters: Filters to query the features data
        :return: Pages object
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.FEATURE)
            filters._user_query = 'false'
        # default sorting
        if filters.sort == dict():
            filters.sort_by(field='id')
        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.FEATURE:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.FEATURE. Got: {!r}'.format(filters.resource))
        if self._feature_set is not None:
            filters.add(field='featureSetId', values=self._feature_set.id)
        if self._item is not None:
            filters.add(field='entityId', values=self._item.id)
        if self._dataset is not None:
            filters.add(field='datasetId', values=self._dataset.id)
        if self._project_id is None:
            self._project_id = self.project.id
        filters.context = {"projects": [self._project_id]}
        
        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    @_api_reference.add(path='/features/vectors/{id}', method='get')
    def get(self, feature_id: str) -> entities.Feature:
        """
        Get Feature object

        :param str feature_id: feature id
        :return: Feature object
        """

        success, response = self._client_api.gen_request(req_type="GET",
                                                         path="{}/{}".format(self.URL, feature_id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Feature.from_json(client_api=self._client_api,
                                          _json=response.json())

    @_api_reference.add(path='/features/vectors', method='post')
    def create(self,
               value,
               project_id: str = None,
               feature_set_id: str = None,
               entity=None,
               version: str = None,
               parent_id: str = None,
               org_id: str = None
               ):
        """
        Create a new Feature vector

        :param immutable value: actual vector - immutable (list of floats [1,2,3]) or list of lists of floats
        :param str project_id: the id of the project where feature will be created
        :param str feature_set_id: ref to a featureSet this vector is a part of
        :param entity: the entity the featureVector is linked to (item, annotation, etc) or list of entities
        :param str version: version of the featureSet generator
        :param str parent_id: optional: parent FeatureSet id - used when FeatureVector is a subFeature
        :param str org_id: the id of the org where featureVector will be created
        :return: Feature vector:
        """
        if project_id is None:
            if self._project is not None:
                project_id = self._project.id
            elif self._project_id is not None:
                project_id = self._project_id
            else:
                raise ValueError('Must insert a project id')

        if feature_set_id is None:
            if self._feature_set is None:
                raise ValueError(
                    'Missing feature_set_id. Must insert the variable or create from context, e.g. feature_set.features.create()')
            feature_set_id = self._feature_set.id
        
        # Additional payload info
        additional_payload = {}
        if version is not None:
            additional_payload['version'] = version
        if parent_id is not None:
            additional_payload['parentId'] = parent_id
        if org_id is not None:
            additional_payload['org'] = org_id
        additional_payload['project'] = project_id
        additional_payload['featureSetId'] = feature_set_id

        if not isinstance(entity, list):
            entity = [entity]
            value = [value]

        if len(value) != len(entity):
            raise ValueError('The number of vectors must be equal to the number of entities')

        payload = []
        for (single_entity, single_value) in zip(entity, value):
            entry = {'entityId': single_entity.id,
                            'value': single_value,
                            'datasetId': single_entity.dataset.id
                        }
            entry.update(additional_payload)
            payload.append(entry)

        success, response = self._client_api.gen_request(req_type="post",
                                                         json_req=payload,
                                                         path=self.URL)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        features = [entities.Feature.from_json(client_api=self._client_api,
                                          _json=feature) for feature in response.json()]
        # return entity
        if len(features) == 1:
            return features[0]
        return features

    @_api_reference.add(path='/features/vectors/{id}', method='delete')
    def delete(self, feature_id: str):
        """
        Delete feature vector

        :param str feature_id: feature id to delete
        :return: success
        :rtype: bool
        """

        success, response = self._client_api.gen_request(req_type="delete",
                                                         path="{}/{}".format(self.URL, feature_id))

        # check response
        if success:
            logger.debug("Feature deleted successfully")
            return success
        else:
            raise exceptions.PlatformException(response)

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Item]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.Feature._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item})
        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items



from collections import namedtuple
import traceback
import logging
import uuid
import attr
import os

from .. import repositories, entities, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


@attr.s
class Recipe(entities.BaseEntity):
    """
    Recipe object
    """
    id = attr.ib()
    creator = attr.ib()
    url = attr.ib(repr=False)
    title = attr.ib()
    project_ids = attr.ib()
    description = attr.ib()
    ontology_ids = attr.ib(repr=False)
    instructions = attr.ib(repr=False)
    examples = attr.ib(repr=False)
    custom_actions = attr.ib(repr=False)
    metadata = attr.ib()
    created_at = attr.ib()
    updated_at = attr.ib()
    updated_by = attr.ib()

    # name change
    ui_settings = attr.ib()

    # platform
    _client_api = attr.ib(type=ApiClient, repr=False)
    # entities
    _dataset = attr.ib(repr=False, default=None)
    _project = attr.ib(repr=False, default=None)
    # repositories
    _repositories = attr.ib(repr=False)

    @property
    def customActions(self):
        return self.custom_actions

    @property
    def ontologyIds(self):
        logger.warning('Deprecation Warning - ontologyIds will be Deprecation from version 1.52.0 use ontology_ids')
        return self.ontology_ids

    @classmethod
    def from_json(cls, _json, client_api, dataset=None, project=None, is_fetched=True):
        """
        Build a Recipe entity object from a json

        :param dict _json: _json response from host
        :param dtlpy.entities.dataset.Dataset dataset: Dataset entity
        :param dtlpy.entities.project.Project project: project entity
        :param dl.ApiClient client_api: ApiClient entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: Recipe object
        """
        project_ids = _json.get('projectIds', None)
        if project is not None and project_ids is not None:
            if project.id not in project_ids:
                logger.warning('Recipe has been fetched from a project that is not belong to it')
                project = None

        inst = cls(
            client_api=client_api,
            dataset=dataset,
            project=project,
            id=_json['id'],
            creator=_json.get('creator', None),
            url=_json.get('url', None),
            title=_json.get('title', None),
            project_ids=project_ids,
            description=_json.get('description', None),
            ontology_ids=_json.get('ontologyIds', None),
            instructions=_json.get('instructions', None),
            ui_settings=_json.get('uiSettings', None),
            metadata=_json.get('metadata', None),
            examples=_json.get('examples', None),
            custom_actions=_json.get('customActions', None),
            created_at=_json.get("createdAt", None),
            updated_at=_json.get("updatedAt", None),
            updated_by=_json.get("updatedBy", None))
        inst.is_fetched = is_fetched
        return inst

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, dataset=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :param project: project entity
        :param dataset:  dataset entity
        :param is_fetched:  is Entity fetched from Platform
        :return:
        """
        try:
            recipe = Recipe.from_json(_json=_json,
                                      client_api=client_api,
                                      project=project,
                                      dataset=dataset,
                                      is_fetched=is_fetched)
            status = True
        except Exception:
            recipe = traceback.format_exc()
            status = False
        return status, recipe

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['ontologies', 'recipes'])
        if self._dataset is None and self._project is None:
            recipes = repositories.Recipes(client_api=self._client_api, dataset=self._dataset, project=self._project)
        elif self._dataset is not None:
            recipes = self.dataset.recipes
        else:
            recipes = self.project.recipes
        r = reps(ontologies=repositories.Ontologies(recipe=self, client_api=self._client_api),
                 recipes=recipes)
        return r

    @property
    def dataset(self):
        if self._dataset is not None:
            assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def project(self):
        if self._project is not None:
            assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Recipe)._client_api,
                                                              attr.fields(Recipe)._dataset,
                                                              attr.fields(Recipe)._project,
                                                              attr.fields(Recipe).project_ids,
                                                              attr.fields(Recipe).ui_settings,
                                                              attr.fields(Recipe)._repositories,
                                                              attr.fields(Recipe).custom_actions,
                                                              attr.fields(Recipe).ontology_ids,
                                                              attr.fields(Recipe).created_at,
                                                              attr.fields(Recipe).updated_at,
                                                              attr.fields(Recipe).updated_by,
                                                              ))
        _json['uiSettings'] = self.ui_settings
        _json['projectIds'] = self.project_ids
        _json['customActions'] = self.custom_actions
        _json['ontologyIds'] = self.ontology_ids
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['updatedBy'] = self.updated_by
        return _json

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/recipes/{}".format(self.project_ids[0], self.id))

    def open_in_web(self):
        """
        Open the recipes in web platform

        :return:
        """
        self._client_api._open_in_web(url=self.platform_url)

    def delete(self, force: bool = False):
        """
        Delete recipe from platform

        :param bool force: force delete recipe

        :return: True
        :rtype: bool
        """
        return self.recipes.delete(self.id, force=force)

    def update(self, system_metadata=False):
        """
        Update Recipe

        :param bool system_metadata: bool - True, if you want to change metadata system
        :return: Recipe object
        :rtype: dtlpy.entities.recipe.Recipe
        """
        return self.recipes.update(recipe=self, system_metadata=system_metadata)

    def clone(self, shallow=False):
        """
        Clone Recipe

       :param bool shallow: If True, link ot existing ontology, clones all ontology that are link to the recipe as well
       :return: Cloned ontology object
       :rtype: dtlpy.entities.recipe.Recipe
        """
        return self.recipes.clone(recipe=self,
                                  shallow=shallow)

    def get_annotation_template_id(self, template_name):
        """
        Get annotation template id by template name

       :param str template_name:
       :return: template id or None if does not exist
        """
        collection_templates = list()
        if 'system' in self.metadata and 'collectionTemplates' in self.metadata['system']:
            collection_templates = self.metadata['system']['collectionTemplates']

        for template in collection_templates:
            if "name" and 'id' in template:
                if template_name == template['name']:
                    return template['id']
        raise exceptions.NotFound('404', "annotation template {!r} not found".format(template_name))

    def add_instruction(self, annotation_instruction_file):
        """
        Add instruction to recipe

        :param str annotation_instruction_file: file path or url of the recipe instruction
        """
        _, ext = os.path.splitext(annotation_instruction_file)
        if ext != '.pdf':
            raise exceptions.PlatformException(error='400',
                                               message='file Must be pdf')
        for project_id in self.project_ids:
            project = repositories.Projects(client_api=self._client_api).get(project_id=project_id)
            dataset = project.datasets._get_binaries_dataset()
            remote_path = '/.dataloop/recipes/{}/instructions'.format(self.id)
            instruction_item = dataset.items.upload(local_path=annotation_instruction_file,
                                                    remote_path=remote_path,
                                                    remote_name=str(uuid.uuid4()) + '.pdf',
                                                    overwrite=True)
            self.metadata['system']['instructionDocument'] = {'itemId': instruction_item.id,
                                                              'datasetId': dataset.id,
                                                              'name': instruction_item.name}
            self.update(True)

    def upload_annotations_verification_file(self, local_path: str, overwrite: bool = False) -> entities.Item:
        """
        Add Annotations Verification js file to the recipe.

        :param str local_path: file path of the annotations verification js file.
        :param bool overwrite: overwrite exiting file if the local and the remote names are matching
        :return: annotations verification js item.
        """

        validation_file_metadata = self.metadata.get("system", dict()).get("validationFile", None)
        if validation_file_metadata is None:
            validation_file_metadata = dict()

        remote_name = validation_file_metadata.get("name", None)
        local_name = os.path.basename(local_path)
        binaries_dataset = self._project.datasets._get_binaries_dataset()
        remote_path = f"/.dataloop/recipes/{self.id}/verification/"

        if remote_name is None or overwrite or remote_name != local_name:
            validation_item = binaries_dataset.items.upload(
                local_path=local_path,
                remote_path=remote_path,
                remote_name=local_name,
                overwrite=True
            )
            self.metadata["system"]["validationFile"] = {
                "itemId": validation_item.id,
                "datasetId": binaries_dataset.id,
                "name": local_name
            }
            self.update(system_metadata=True)
        else:
            logger.debug(f"Existing Annotations Validation Script was found.")
            validation_item_id = self.metadata["system"]["validationFile"]["itemId"]
            validation_item = binaries_dataset.items.get(item_id=validation_item_id)
        return validation_item



import hashlib
import logging
import os
import io
import random
from typing import List

from .. import entities, PlatformException, exceptions, repositories, miscellaneous
from ..services.api_client import ApiClient
from ..services.api_client import client as client_api

logger = logging.getLogger(name='dtlpy')


class Codebases:
    """
    Codebase Repository

    The Codebases class allows the user to manage codebases and their properties.
    The codebase is the code the user uploads for the user's packages to run.
    Read more about `codebase <https://developers.dataloop.ai/tutorials/faas/introduction/chapter/>`_ in our FaaS (function as a service).
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 dataset: entities.Dataset = None,
                 project_id: str = None):
        self._client_api = client_api
        self._project = project
        self._project_id = project_id
        self._dataset = dataset
        self._items_repository = None
        self.git_utils = miscellaneous.GitUtils()

    @property
    def items_repository(self) -> repositories.Items:
        if self._items_repository is None:
            if self._dataset is not None or self._project is not None:
                self._items_repository = self.dataset.items
            else:
                self._items_repository = repositories.Items(client_api=client_api)
        assert isinstance(self._items_repository, repositories.Items)
        return self._items_repository

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # project is None - try dataset
            if self._dataset is None:
                # dataset is None - try from project id
                if self._project_id is None:
                    raise PlatformException(error='400',
                                            message='Cannot perform this action without a project')
                else:
                    self._project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)
            else:
                # dataset is not None - take project from there
                self._project = self.dataset.project
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            # get dataset from project
            try:
                self._dataset = self.project.datasets._get_binaries_dataset()
            except exceptions.NotFound:
                raise ValueError('Missing "Binaries" dataset in the project. Please contact support for help')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    @staticmethod
    def __file_hash(filepath):
        m = hashlib.md5()
        with open(filepath, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                m.update(chunk)
        return m.hexdigest()

    def list_versions(self, codebase_name: str) -> entities.PagedEntities:
        """
        List all codebase versions.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        **Example**:

        .. code-block:: python

            package.codebases.list_versions(codebase_name='codebase_name')

        :param str codebase_name: code base name
        :return: list of versions
        :rtype: list
        """
        filters = entities.Filters()
        filters.add(field='filename', values='/codebases/{}/*'.format(codebase_name))
        versions = self.items_repository.list(filters=filters)
        return versions

    def list(self) -> entities.PagedEntities:
        """
        List all codebases.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        **Example**:

        .. code-block:: python

            package.codebases.list()

        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        filters = entities.Filters()
        filters.add(field='filename', values='/codebases/*')
        filters.add(field='type', values='dir')
        codebases = self.items_repository.list(filters=filters)
        return codebases

    def get(self,
            codebase_name: str = None,
            codebase_id: str = None,
            version: str = None):
        """
        Get a Codebase object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        **Example**:

        .. code-block:: python

            package.codebases.get(codebase_name='codebase_name')

        :param str codebase_name: optional - search by name
        :param str codebase_id: optional - search by id
        :param str version: codebase version. default is latest. options: "all", "latest" or ver number - "10"
        :return: Codebase object
        """
        if codebase_id is not None:
            matched_version = self.items_repository.get(item_id=codebase_id)
            # verify input codebase name is same as the given id
            if codebase_name is not None and matched_version.name != codebase_name:
                logger.warning(
                    "Mismatch found in codebases.get: codebase_name is different then codebase.name: "
                    "{!r} != {!r}".format(
                        codebase_name,
                        matched_version.name))
            codebase = entities.Codebase(type='item',
                                         client_api=self._client_api,
                                         itemId=matched_version.id,
                                         item=matched_version)
            return codebase

        if codebase_name is None:
            raise PlatformException(error='400', message='Either "codebase_name" or "codebase_id" is needed')
        if version is None:
            version = 'latest'

        if version not in ['all', 'latest']:
            try:
                matched_version = self.items_repository.get(
                    filepath='/codebases/{}/{}.zip'.format(codebase_name, version))
            except Exception:
                raise PlatformException(error='404',
                                        message='No matching version was found. version: {}'.format(version))
            codebase = entities.Codebase(type='item',
                                         client_api=self._client_api,
                                         itemId=matched_version.id,
                                         item=matched_version)
            return codebase

        # get all or latest
        versions_pages = self.list_versions(codebase_name=codebase_name)
        if versions_pages.items_count == 0:
            raise PlatformException(error='404', message='No codebase was found. name: {}'.format(codebase_name))
        else:
            if version == 'all':
                codebase = [entities.Codebase(type='item',
                                              client_api=self._client_api,
                                              itemId=mv.id,
                                              item=mv) for mv in versions_pages.all()]
            elif version == 'latest':
                max_ver = -1
                matched_version = None
                for page in versions_pages:
                    for ver in page:
                        if ver.type == 'dir':
                            continue
                        # extract version from filepath
                        ver_int = int(os.path.splitext(ver.name)[0])
                        if ver_int > max_ver:
                            max_ver = ver_int
                            matched_version = ver
                if matched_version is None:
                    raise PlatformException(error='404',
                                            message='No codebase was found. name: {}'.format(codebase_name))
                else:
                    codebase = entities.Codebase(type='item',
                                                 client_api=self._client_api,
                                                 itemId=matched_version.id,
                                                 item=matched_version)
            else:
                raise PlatformException(error='404', message='Unknown version string: {}'.format(version))

        return codebase

    @staticmethod
    def get_current_version(all_versions_pages, zip_md):
        """
        This method returns the current version of the codebase and other versions found.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param codebase all_versions_pages: codebase object
        :param zip_md: zipped file of codebase
        :return: current version and all versions found of codebase
        :rtype: int, int

        **Example**:

        .. code-block:: python

            package.codebases.get_current_version(all_versions_pages='codebase_entity', zip_md='path')
        """
        latest_version = 0
        same_version_found = None
        # go over all existing versions
        for v_item in all_versions_pages:
            # get latest version
            if int(os.path.splitext(v_item.item.name)[0]) > latest_version:
                latest_version = int(os.path.splitext(v_item.item.name)[0])
            # check md5 to find same codebase
            if 'md5' in v_item.item.metadata['system'] and v_item.item.metadata['system']['md5'] == zip_md:
                same_version_found = v_item
                break
        return latest_version + 1, same_version_found

    def pack(self,
             directory: str,
             name: str = None,
             extension: str = 'zip',
             description: str = '',
             ignore_directories: List[str] = None,
             ignore_max_file_size: bool = False):
        """
        Zip a local code directory and post to codebases.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param str directory: local directory to pack
        :param str name: codebase name
        :param str extension: the extension of the file
        :param str description: codebase description
        :param list[str] ignore_directories: directories to ignore.
        :return: Codebase object
        :rtype: dtlpy.entities.codebase.Codebase

        **Example**:

        .. code-block:: python

            package.codebases.pack(directory='path_dir', name='codebase_name')
        """
        # create/get .dataloop dir
        cwd = os.getcwd()
        dl_dir = os.path.join(cwd, '.dataloop')
        if not os.path.isdir(dl_dir):
            os.mkdir(dl_dir)

        # get codebase name
        if name is None:
            name = os.path.basename(directory)

        # create/get dist folder
        zip_filename = os.path.join(dl_dir, '{}_{}.{}'.format(name, str(random.randrange(0, 1000)), extension))

        try:
            if not os.path.isdir(directory):
                raise PlatformException(error='400', message='Not a directory: {}'.format(directory))
            directory = os.path.abspath(directory)

            # create zipfile
            miscellaneous.Zipping.zip_directory(zip_filename=zip_filename,
                                                directory=directory,
                                                ignore_directories=ignore_directories,
                                                ignore_max_file_size=ignore_max_file_size)
            zip_md = self.__file_hash(zip_filename)

            # get latest version
            same_version_found = None
            try:
                all_versions_pages = self.get(codebase_name=name, version='all')
            except exceptions.NotFound:
                all_versions_pages = None
            if all_versions_pages is None:
                # no codebase with that name - create new version
                current_version = 0
            else:
                current_version, same_version_found = self.get_current_version(all_versions_pages=all_versions_pages,
                                                                               zip_md=zip_md)

            if same_version_found is not None:
                # same md5 hash file found in version - return the matched version
                codebase = same_version_found
            else:
                # no matched version was found - create a new version
                # read from zipped file
                with open(zip_filename, 'rb') as f:
                    buffer = io.BytesIO(f.read())
                    buffer.name = '{}.{}'.format(str(current_version), extension)

                # upload item
                item = self.items_repository.upload(local_path=buffer,
                                                    remote_path='/codebases/{}'.format(name))
                if isinstance(item, list) and len(item) == 0:
                    raise PlatformException(error='400', message='Failed upload codebase, check log file for details')

                # add source code to metadata
                if 'system' not in item.metadata:
                    item.metadata['system'] = dict()
                item.metadata['system']['description'] = description
                item.metadata['system']['md5'] = zip_md

                # add git info to metadata
                if miscellaneous.GitUtils.is_git_repo(path=directory):
                    # create 'git' field in metadata
                    if 'git' not in item.metadata:
                        item.metadata['git'] = dict()

                    # add to metadata
                    item.metadata['git']['status'] = miscellaneous.GitUtils.git_status(path=directory)
                    item.metadata['git']['log'] = miscellaneous.GitUtils.git_log(path=directory)
                    item.metadata['git']['url'] = miscellaneous.GitUtils.git_url(path=directory)

                # update item
                item = self.items_repository.update(item=item, system_metadata=True)
                codebase = entities.Codebase(type='item',
                                             client_api=self._client_api,
                                             itemId=item.id,
                                             item=item)
        except Exception:
            logger.error('Error when packing:')
            raise
        finally:
            # cleanup
            if zip_filename is not None:
                if os.path.isfile(zip_filename):
                    os.remove(zip_filename)
        return codebase

    def _unpack_single(self,
                       codebase,
                       download_path: str,
                       local_path: str):
        """
        :param dtlpy.entities.codebase.Codebase codebase: codebase object
        :param str download_path:
        :param str local_path:
        """
        # downloading with specific filename
        if isinstance(codebase, entities.ItemCodebase):
            artifact_filepath = self.items_repository.download(items=codebase.item_id,
                                                               save_locally=True,
                                                               local_path=os.path.join(download_path,
                                                                                       codebase.item.name),
                                                               to_items_folder=False)
            if not os.path.isfile(artifact_filepath):
                raise PlatformException(error='404',
                                        message='error downloading codebase. see above for more information')
            miscellaneous.Zipping.unzip_directory(zip_filename=artifact_filepath,
                                                  to_directory=local_path)
            os.remove(artifact_filepath)
            logger.info('Source code was unpacked to: {}'.format(artifact_filepath))
        elif isinstance(codebase, entities.Item):
            artifact_filepath = codebase.download(save_locally=True,
                                                  local_path=os.path.join(download_path,
                                                                          codebase.name),
                                                  to_items_folder=False)
            if not os.path.isfile(artifact_filepath):
                raise PlatformException(error='404',
                                        message='error downloading codebase. see above for more information')
            miscellaneous.Zipping.unzip_directory(zip_filename=artifact_filepath,
                                                  to_directory=local_path)
            os.remove(artifact_filepath)
            logger.info('Source code was unpacked to: {}'.format(artifact_filepath))
        elif isinstance(codebase, entities.GitCodebase):
            if codebase.is_git_repo(local_path) or \
                    codebase.is_git_repo(os.path.join(local_path, codebase.git_repo_name)):
                artifact_filepath = self.pull_git(codebase=codebase, local_path=local_path)
            else:  # Clone the repo if not exist
                artifact_filepath = self.clone_git(codebase=codebase, local_path=local_path)
        else:
            raise ValueError('Not implemented: "_unpack_single" for codebase type: {!r}'.format(codebase.type))
        return artifact_filepath

    def _get_single(self, codebase: entities.GitCodebase, name: str):
        value = None
        integration_id = codebase.credentials.get(name, {}).get('id', None)
        key = codebase.credentials.get(name, {}).get('key', None)
        if integration_id is not None:
            try:
                key = self.project.integrations.get(integrations_id=integration_id).name
            except Exception:
                pass
            value = os.environ.get(key, None)

        return value

    def _get_credential(self, codebase: entities.GitCodebase):
        username = password = None

        if codebase.credentials:
            try:
                username = self._get_single(codebase=codebase, name='username')
                password = self._get_single(codebase=codebase, name='password')
            except Exception:
                logger.exception('Failed to get credentials from codebase')

        return username, password

    def clone_git(self,
                  codebase: entities.Codebase,
                  local_path: str):
        """
        Clone code base

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.codebase.Codebase codebase: codebase object
        :param str local_path: local path
        :return: path where the clone will be
        :rtype: str

         **Example**:

        .. code-block:: python

            package.codebases.clone_git(codebase='codebase_entity', local_path='local_path')
        """
        if not isinstance(codebase, entities.GitCodebase):
            raise RuntimeError('only support Git Codebase')
        username, password = self._get_credential(codebase=codebase)
        response = self.git_utils.git_clone(path=local_path,
                                            git_url=codebase.git_url,
                                            tag=codebase.git_tag,
                                            username=username,
                                            password=password)
        if response:
            logger.info('Source code was cloned from {}(Git) to: {}'.format(codebase.git_url, local_path))
        else:
            raise RuntimeError('Failed cloning. See above for full log. codebase: {}'.format(codebase))
        return local_path

    def pull_git(self, codebase, local_path):
        """
        Pull (download) a codebase.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.codebase.Codebase codebase: codebase object
        :param str local_path: local path
        :return: path where the Pull will be
        :rtype: str

        **Example**:

        .. code-block:: python

            package.codebases.pull_git(codebase='codebase_entity', local_path='local_path')
        """
        pull_cmd = 'git pull'
        if not codebase.is_git_repo(local_path):
            local_path = os.path.join(local_path, codebase.git_repo_name)
        response = self.git_utils.git_command(path=local_path, cmd=pull_cmd)
        if response:
            logger.info('pull successful {}(Git) to: {}'.format(codebase.git_url, os.path.dirname(local_path)))
        else:
            logger.critical("Could not pull")

        # we can test if this is not the same repo if needed...
        # FIXME need to change the order - checkout new branch and pull
        response_2 = self.git_utils.git_command(path=local_path, cmd='git checkout {}'.format(codebase.git_tag))
        return local_path

    def unpack(self,
               codebase: entities.Codebase = None,
               codebase_name: str = None,
               codebase_id: str = None,
               local_path: str = None,
               version: str = None):
        """
        Unpack codebase locally. Download source code and unzip.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a package.

        :param dtlpy.entities.codebase.Codebase codebase: `dl.Codebase` object
        :param str codebase_name: search by name
        :param str codebase_id: search by id
        :param str local_path: local path to save codebase
        :param str version: codebase version to unpack. default - latest
        :return: String (dirpath)
        :rtype: str

        **Example**:

        .. code-block:: python

            package.codebases.unpack(codebase='codebase_entity', local_path='local_path')
        """
        # get the codebase / multiple codebase
        if codebase is None:
            codebase = self.get(codebase_name=codebase_name,
                                codebase_id=codebase_id,
                                version=version)
        elif codebase_name is not None or codebase_id is not None:
            logger.warning("Using given codebase. Does not preforming search with name {!r} / id {!r}".
                           format(codebase_name, codebase_id))
        download_path = local_path
        if isinstance(codebase, entities.PagedEntities):
            for page in codebase:
                for item in page:
                    local_path = os.path.join(download_path, 'v.' + item.name.split('.')[0])
                    self._unpack_single(codebase=item,
                                        download_path=download_path,
                                        local_path=local_path)
            return os.path.dirname(local_path)
        elif isinstance(codebase, list):
            for item in codebase:
                local_path = os.path.join(download_path, 'v.' + item.item.name.split('.')[0])
                self._unpack_single(codebase=item,
                                    download_path=download_path,
                                    local_path=local_path)
            return os.path.dirname(local_path)
        elif isinstance(codebase, (entities.ItemCodebase, entities.Item, entities.GitCodebase)):
            artifact_filepath = self._unpack_single(codebase=codebase,
                                                    download_path=download_path,
                                                    local_path=local_path)
            if isinstance(codebase, (entities.ItemCodebase, entities.Item)):
                dir_path = os.path.dirname(artifact_filepath)  # use the directory of the artifact
            else:
                dir_path = artifact_filepath
            logger.info('Source code was unpacked to: {}'.format(dir_path))
        else:
            raise PlatformException(
                error='404',
                message='Codebase was not found! name:{name}, id:{id}'.format(name=codebase_name,
                                                                              id=codebase_id))
        return dir_path



import logging
from .dataset_generator import DatasetGenerator

logger = logging.getLogger(name='dtlpy')

try:
    import torch
    from torch.utils.data import Dataset
except (ImportError, ModuleNotFoundError):
    logger.error('Failed importing torch package, cannot use DatasetGeneratorTorch')


class DatasetGeneratorTorch(DatasetGenerator, Dataset):
    """

    """

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        return super(DatasetGeneratorTorch, self).__getitem__(idx)



import logging
from urllib.parse import urlencode
from .. import exceptions, entities, miscellaneous
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Webhooks:
    """
    Webhooks repository
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None):
        self._project = project
        self._client_api = client_api
        self._url = '/webhooks'

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.webhooks repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def create(self,
               name: str,
               project_id: str = None,
               http_method=None,
               hook_url=None,
               project: entities.Project = None) -> entities.Webhook:
        """
        Create web hook entity

        :param name:
        :param project_id:
        :param http_method:
        :param hook_url:
        :param project:
        :return:
        """
        if project is None:
            project = self._project

        if project_id is None and project is None:
            raise exceptions.PlatformException('400', 'Must provide project or project id')

        if project_id is None:
            project_id = project.id

        # payload
        payload = {
            'name': name,
            'httpMethod': http_method,
            'hookUrl': hook_url,
            'project': project_id
        }

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=self._url,
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Webhook.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=project)

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List web hooks

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.WEBHOOK)
            if self._project is not None:
                filters.add(field='projectId', values=self._project.id)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))

        if filters.resource != entities.FiltersResource.WEBHOOK:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.WEBHOOK. Got: {!r}'.format(filters.resource))

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Webhook]:
        # handle execution
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return execution list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.Webhook.from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'project': self._project})

        # get all results
        items = miscellaneous.List([j.result() for j in jobs])
        return items

    def _list(self, filters: entities.Filters):
        """
        List web hooks

        :return:
        """

        query_params = {
            'pageOffset': filters.page,
            'pageSize': filters.page_size
        }

        url = self._url + '?{}'.format(
            urlencode({key: val for key, val in query_params.items() if val is not None}, doseq=True))

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)

        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def get(self, webhook_id=None, webhook_name=None) -> entities.Webhook:
        """
        Get Web hook

        :param webhook_id:
        :param webhook_name:
        :return: Web hook execution object
        """
        if webhook_id is None and webhook_name is None:
            raise exceptions.PlatformException('400', 'Must provide webhook_id or webhook_name id')
        elif webhook_id is None:
            webhooks_pages = self.list(
                filters=entities.Filters(field='name', values=webhook_name, resource=entities.FiltersResource.WEBHOOK))
            if webhooks_pages.items_count == 0:
                raise exceptions.PlatformException('404', 'Not found: web hook: {}'.format(webhook_name))
            elif webhooks_pages.items_count > 1:
                raise exceptions.PlatformException('404',
                                                   'More than one webhooks found: web hook: {}'.format(webhook_name))
            else:
                webhook = webhooks_pages.items[0]
        else:
            success, response = self._client_api.gen_request(
                req_type="get",
                path="{}/{}".format(self._url, webhook_id)
            )

            # exception handling
            if not success:
                raise exceptions.PlatformException(response)

            # return entity
            webhook = entities.Webhook.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 project=self._project)
            # verify input webhook name is same as the given id
            if webhook_name is not None and webhook.name != webhook_name:
                logger.warning(
                    "Mismatch found in webhooks.get: webhook_name is different then webhook.name:"
                    " {!r} != {!r}".format(
                        webhook_name,
                        webhook.name))

        return webhook

    def delete(self, webhook_id=None, webhook_name=None):
        """
        Delete Trigger object

        :param webhook_id:
        :param webhook_name:
        :return: True
        """
        if webhook_id is None:
            if webhook_name is None:
                raise exceptions.PlatformException('400', 'Must provide either webhook name or webhook id')
            else:
                webhook_id = self.get(webhook_name=webhook_name).id

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="{}/{}".format(self._url, webhook_id)
        )
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        return True

    def update(self, webhook: entities.Webhook) -> entities.Webhook:
        """

        :param webhook: Webhook entity
        :return: Webhook entity
        """
        assert isinstance(webhook, entities.Webhook)

        # payload
        payload = webhook.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='{}/{}'.format(self._url, webhook.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Webhook.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=self._project)



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.
from .annotations import Annotations
from .artifacts import Artifacts
from .datasets import Datasets
from .items import Items
from .codebases import Codebases
from .projects import Projects
from .ontologies import Ontologies
from .recipes import Recipes
from .packages import Packages, PackageCatalog
from .downloader import Downloader
from .uploader import Uploader
from .triggers import Triggers
from .times_series import TimesSeries
from .services import Services, FUNCTION_END_LINE
from .executions import Executions
from .assignments import Assignments
from .tasks import Tasks
from .bots import Bots
from .webhooks import Webhooks
from .models import Models, Metrics
from .analytics import Analytics
from .drivers import Drivers
from .commands import Commands
from .pipelines import Pipelines
from .nodes import Nodes
from .pipeline_executions import PipelineExecutions
from .features import Features
from .feature_sets import FeatureSets
from .organizations import Organizations
from .integrations import Integrations
from .drivers import Drivers
from .settings import Settings
from .resource_executions import ResourceExecutions
from .apps import Apps
from .dpks import Dpks
from .messages import Messages
from .compositions import Compositions
from .schema import Schema
from .computes import Computes
from .service_drivers import ServiceDrivers
from .collections import Collections



from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import History
from fuzzyfinder.main import fuzzyfinder
import csv
import threading
import os
import datetime
import logging


class StateEnum:
    """
    State enum
    """
    START = 0
    RUNNING = 1
    DONE = 2
    CONTINUE = 3


def get_parser_tree(parser):
    """
    Creates parser tree for autocomplete
    :param parser: parser
    :return: parser tree
    """
    if parser._subparsers is None:
        p_keywords = list()
        for param in parser._option_string_actions:
            if param.startswith('--'):
                p_keywords.append(param)
    else:
        p_keywords = dict()
        subparsers = parser._subparsers._group_actions[0].choices
        for sub_parser in subparsers:
            p_keywords[sub_parser] = get_parser_tree(subparsers[sub_parser])
    if 'shell' in p_keywords:
        # noinspection PyTypeChecker
        p_keywords.pop('shell')
    return p_keywords


class FileHistory(History):
    """
    :class:`.History` class that stores all strings in a file.
    """

    def __init__(self, filename):
        self.filename = filename
        super(FileHistory, self).__init__()
        self.to_hide = ['password', 'secret']

    def load_history_strings(self):
        strings = []
        lines = []

        def add():
            if lines:
                # Join and drop trailing newline.
                string = ''.join(lines)[:-1]
                hide = any(field in string for field in self.to_hide)
                if not hide:
                    strings.append(string)

        if os.path.exists(self.filename):
            with open(self.filename, 'rb') as f:
                for line in f:
                    line = line.decode('utf-8')

                    if line.startswith('+'):
                        lines.append(line[1:])
                    else:
                        add()
                        lines = []

                add()

        # Reverse the order, because newest items have to go first.
        return reversed(strings)

    def store_string(self, string):
        # Save to file.
        with open(self.filename, 'ab') as f:
            def write(t):
                f.write(t.encode('utf-8'))

            write('\n# %s\n' % datetime.datetime.now(datetime.timezone.utc))
            for line in string.split('\n'):
                hide = any(field in line for field in self.to_hide)
                if not hide:
                    write('+%s\n' % line)


class DlpCompleter(Completer):
    """
    Autocomplete for dlp shell
    """

    def __init__(self, keywords, dlp):
        super(DlpCompleter, self).__init__()
        # globals
        self.keywords = keywords
        self.param_suggestions = list()
        self.thread_state = StateEnum.START
        self.dlp = dlp

    def get_param_suggestions(self, param, word_before_cursor, cmd):
        """
        Return parap suggestions
        :param param:
        :param word_before_cursor:
        :param cmd:
        :return:
        """
        prev_state = logging.root.manager.disable
        try:
            logging.root.manager.disable = logging.ERROR
            if self.thread_state in [StateEnum.RUNNING, StateEnum.DONE]:
                return
            else:
                if param == '--project-name':
                    self.thread_state = StateEnum.RUNNING
                    project_list = self.dlp.projects.list()
                    self.param_suggestions = ['"{}'.format(project.name) for project in project_list]
                    self.thread_state = StateEnum.DONE

                elif param == '--package-name':
                    self.thread_state = StateEnum.RUNNING
                    if '--project-name' in cmd:
                        project = self.dlp.projects.get(
                            project_name=cmd[cmd.index('--project-name') + 1].replace('"', ''))
                        packages = project.packages
                    else:
                        packages = self.dlp.packages
                    package_list = packages.list()
                    self.param_suggestions = ['"{}'.format(package.name) for package in package_list.items]
                    self.thread_state = StateEnum.DONE

                elif param == '--service-name':
                    self.thread_state = StateEnum.RUNNING
                    if '--package-name' in cmd:
                        package = self.dlp.packages.get(
                            package_name=cmd[cmd.index('--package-name') + 1].replace('"', ''))
                        services = package.services
                    elif '--project-name' in cmd:
                        project = self.dlp.projects.get(
                            project_name=cmd[cmd.index('--project-name') + 1].replace('"', ''))
                        services = project.services
                    else:
                        services = self.dlp.services

                    service_list = services.list()
                    self.param_suggestions = ['"{}'.format(service.name) for service in service_list.items]
                    self.thread_state = StateEnum.DONE

                elif param == '--trigger-name':
                    self.thread_state = StateEnum.RUNNING
                    if '--service-name' in cmd:
                        service = self.dlp.services.get(
                            service_name=cmd[cmd.index('--service-name') + 1].replace('"', ''))
                        triggers = service.triggers
                    elif '--package-name' in cmd:
                        package = self.dlp.packages.get(
                            package_name=cmd[cmd.index('--package-name') + 1].replace('"', ''))
                        triggers = package.services
                    elif '--project-name' in cmd:
                        project = self.dlp.projects.get(
                            project_name=cmd[cmd.index('--project-name') + 1].replace('"', ''))
                        triggers = project.services
                    else:
                        triggers = self.dlp.services

                    trigger_list = triggers.list()
                    self.param_suggestions = ['"{}'.format(trigger.name) for trigger in trigger_list.items]
                    self.thread_state = StateEnum.DONE

                elif param == '--dataset-name':
                    self.thread_state = StateEnum.RUNNING
                    if '--project-name' in cmd:
                        project = self.dlp.projects.get(
                            project_name=cmd[cmd.index('--project-name') + 1].replace('"', ''))
                        dataset_list = project.datasets.list()
                    else:
                        project = self.dlp.projects.get()
                        dataset_list = project.datasets.list()
                    self.param_suggestions = ['"{}'.format(dataset.name) for dataset in dataset_list]
                    self.thread_state = StateEnum.DONE

                elif param == '--remote-path':
                    self.thread_state = StateEnum.RUNNING
                    if '--project-name' in cmd:
                        project = self.dlp.projects.get(
                            project_name=cmd[cmd.index('--project-name') + 1].replace('"', ''))
                    else:
                        project = self.dlp.projects.get()
                    if '--dataset-name' in cmd:
                        dataset = project.datasets.get(
                            dataset_name=cmd[cmd.index('--dataset-name') + 1].replace('"', ''))
                    else:
                        dataset = self.dlp.datasets.get()
                    self.param_suggestions = dataset.directory_tree.dir_names
                    with_quotation = ['"{}'.format(directory) for directory in dataset.directory_tree.dir_names]
                    self.param_suggestions += with_quotation
                    self.thread_state = StateEnum.DONE

                elif param in ['--local-path', '--local-annotations-path', '--service-file']:
                    self.thread_state = StateEnum.CONTINUE
                    param = word_before_cursor.replace('"', '')
                    if param == '':
                        param, path = os.path.splitdrive(os.getcwd())
                        param += os.path.sep
                        self.param_suggestions = ['"{}'.format(os.path.join(param, directory))
                                                  for directory in os.listdir(param)
                                                  if not directory.startswith('.')]
                    elif os.path.isdir(os.path.dirname(param)):
                        base_dir = os.path.dirname(param)
                        self.param_suggestions = ['"{}'.format(os.path.join(base_dir, directory))
                                                  for directory in os.listdir(base_dir)
                                                  if (not directory.startswith('.') and
                                                      param in '"{}'.format(os.path.join(base_dir, directory)))]

                elif param in ['--annotation-options']:
                    self.thread_state = StateEnum.CONTINUE
                    self.param_suggestions = ['mask', 'json', 'instance', '"mask, json"',
                                              '"mask, instance"', '"json, instance"', '"mask, json, instance"']

                elif param in ['--annotation-filter-type']:
                    self.thread_state = StateEnum.CONTINUE
                    self.param_suggestions = ['box', 'segment', 'binary', '"box, segment"']

                elif param in ['--annotation-filter-label']:
                    self.thread_state = StateEnum.CONTINUE

                elif param in ['--actions']:
                    self.thread_state = StateEnum.CONTINUE
                    self.param_suggestions = ['Created', 'Updated', 'Deleted', '"Created, Updated"',
                                              '"Created, Deleted"', '"Updated, Deleted"', '"Created, Updated, Deleted"']

                elif param in ['--resource']:
                    self.thread_state = StateEnum.CONTINUE
                    self.param_suggestions = [value for key, value in self.dlp.TriggerResource.__dict__.items() if
                                              not key.startswith('_')]

                elif param in ['--package-type']:
                    self.thread_state = StateEnum.CONTINUE
                    self.param_suggestions = [value for key, value in self.dlp.PackageCatalog.__dict__.items() if
                                              not key.startswith('_')]

                # elif param in ['cd']:
                #     self.thread_state = StateEnum.CONTINUE
                #     if word_before_cursor != '' and os.path.isdir(os.path.join(os.getcwd(), word_before_cursor)):
                #         self.param_suggestions = [os.path.join(word_before_cursor, directory)
                #                                   for directory in
                #                                   os.listdir(os.path.join(os.getcwd(), word_before_cursor))
                #                                   if os.path.isdir(
                #                 os.path.join(os.getcwd(), word_before_cursor, directory))]
                #     else:
                #         self.param_suggestions = [directory for directory in os.listdir(os.getcwd()) if
                #                                   os.path.isdir(directory)]

                else:
                    self.thread_state = StateEnum.START
                    self.param_suggestions = list()
        except Exception:
            self.param_suggestions = list()
            self.thread_state = StateEnum.START
        finally:
            logging.root.manager.disable = prev_state

    def need_param(self, cmd, word_before_cursor):
        """
        :param cmd:
        :param word_before_cursor:
        """
        need_param = False

        try:
            bool_flags_list = ['--overwrite', '--with-text', '--deploy', '--not-items-folder', '--encode',
                               '--without-binaries']

            if len(cmd) > 2 and cmd[-2].startswith('--') and word_before_cursor != '':
                need_param = self.get_param(cmd=cmd, word_before_cursor=word_before_cursor) not in bool_flags_list
            elif cmd[-1].startswith('-') and word_before_cursor == '':
                need_param = self.get_param(cmd=cmd, word_before_cursor=word_before_cursor) not in bool_flags_list
            elif cmd[0] in ['cd']:
                need_param = True
        except Exception:
            need_param = False

        return need_param

    @staticmethod
    def get_param(cmd, word_before_cursor):
        """
        :param cmd:
        :param word_before_cursor:
        """
        if word_before_cursor == '' or len(cmd) < 2:
            param = cmd[-1]
        else:
            param = cmd[-2]

        return param

    def get_completions(self, document, complete_event):
        """
        Get command completion
        :param document:
        :param complete_event:
        :return:
        """
        # fix input
        cmd = next(csv.reader([" ".join(document.text.split())], delimiter=' '))

        # get current word
        word_before_cursor = document.get_word_before_cursor(WORD=True)

        # suggest keywords
        suggestions = list()
        if self.need_param(cmd=cmd, word_before_cursor=word_before_cursor):
            param = self.get_param(cmd=cmd, word_before_cursor=word_before_cursor)
            if self.thread_state in [StateEnum.START, StateEnum.CONTINUE]:
                if param in ['--project-name',
                             '--dataset-name',
                             '--remote-path',
                             '--package-name',
                             '--service-name',
                             '--trigger-name']:
                    thread = threading.Thread(target=self.get_param_suggestions,
                                              kwargs={"param": param,
                                                      'word_before_cursor': word_before_cursor,
                                                      'cmd': cmd})
                    thread.daemon = True
                    thread.start()

                else:
                    self.get_param_suggestions(param=param, word_before_cursor=word_before_cursor, cmd=cmd)
            if self.thread_state in [StateEnum.DONE, StateEnum.CONTINUE]:
                suggestions = self.param_suggestions
        elif len(cmd) == 0:
            suggestions = list(self.keywords.keys())
        elif len(cmd) == 1:
            self.thread_state = StateEnum.START
            if cmd[0] not in self.keywords.keys() and cmd[0] != '':
                if not word_before_cursor == '':
                    suggestions = list(self.keywords.keys())
            elif cmd[0] == '':
                suggestions = list(self.keywords.keys())
            elif isinstance(self.keywords[cmd[0]], list):
                suggestions = self.keywords[cmd[0]]
            elif isinstance(self.keywords[cmd[0]], dict):
                suggestions = list(self.keywords[cmd[0]].keys())
        elif len(cmd) >= 2:
            self.thread_state = StateEnum.START
            if cmd[0] not in self.keywords.keys():
                suggestions = list()
            elif isinstance(self.keywords[cmd[0]], list):
                suggestions = self.keywords[cmd[0]]
            elif isinstance(self.keywords[cmd[0]], dict):
                if cmd[1] in self.keywords[cmd[0]].keys():
                    suggestions = self.keywords[cmd[0]][cmd[1]]
                else:
                    suggestions = list(self.keywords[cmd[0]].keys())

        matches = fuzzyfinder(word_before_cursor, suggestions)
        for match in matches:
            yield Completion(match, start_position=-len(word_before_cursor))



class Dtlpy:
    from .services.api_client import client as client_api
    from .services.api_client import VerboseLoggingLevel, ApiClient
    from .services import DataloopLogger, DtlpyFilter, check_sdk, Reporter, service_defaults
    from .services.api_reference import api_reference as _api_reference
    from .caches.cache import CacheConfig, CacheType
    from .exceptions import PlatformException
    from . import repositories, exceptions, entities, examples
    from .entities import (
        # main entities
        Project, Dataset, ExpirationOptions, ExportVersion, Trigger, Item, Execution, AnnotationCollection, Annotation,
        Recipe, IndexDriver, AttributesTypes, AttributesRange, Dpk, App, AppModule, AppScope,
        Ontology, Label, Task, TaskPriority, ConsensusTaskType, AllocationMethod, Assignment, Service, Package, Codebase, Model,
        PackageModule, PackageFunction,
        # annotations
        Box, Cube, Cube3d, Point, Note, Message, Segmentation, Ellipse, Classification, Subtitle, Polyline, Pose, Gis, GisType,
        Description,
        Polygon, Text, FreeText, RefImage,
        # filters
        Filters, FiltersKnownFields, FiltersResource, FiltersOperations, FiltersMethod, FiltersOrderByDirection,
        FiltersKnownFields as KnownFields,
        # triggers
        TriggerResource, TriggerAction, TriggerExecutionMode, TriggerType,
        # faas
        FunctionIO, KubernetesAutoscalerType, KubernetesRabbitmqAutoscaler, KubernetesAutoscaler, KubernetesRuntime,
        InstanceCatalog, PackageInputType, ServiceType, ServiceModeType,
        PackageSlot, SlotPostAction, SlotPostActionType, SlotDisplayScope, SlotDisplayScopeResource, UiBindingPanel,
        # roberto
        DatasetSubsetType, ModelStatus, PlotSample, ArtifactType, Artifact, ItemArtifact, LinkArtifact, LocalArtifact,
        EntityScopeLevel,
        # features
        FeatureEntityType, Feature, FeatureSet,
        #
        RequirementOperator, PackageRequirement,
        Command, CommandsStatus,
        LocalCodebase, GitCodebase, ItemCodebase, FilesystemCodebase, PackageCodebaseType,
        MemberRole, MemberOrgRole,
        Webhook, HttpMethod,
        ViewAnnotationOptions, AnnotationStatus, AnnotationType,
        ItemStatus, ExecutionStatus, ExportMetadata,
        PromptItem, Prompt, PromptType, ItemGis, Layer,
        ItemLink, UrlLink, LinkTypeEnum,
        Modality, ModalityTypeEnum, ModalityRefTypeEnum,
        Workload, WorkloadUnit, ItemAction,
        PipelineExecution, CycleRerunMethod, PipelineExecutionNode, Pipeline, PipelineConnection,
        PipelineNode, TaskNode, CodeNode, PipelineStats, PipelineSettings,
        PipelineNodeType, PipelineNameSpace, PipelineResumeOption, Variable, CompositionStatus,
        FunctionNode, DatasetNode, PipelineConnectionPort, PipelineNodeIO, Organization, OrganizationsPlans,
        Integration,
        Driver, S3Driver, GcsDriver, AzureBlobDriver, CacheAction, PodType,
        ExternalStorage, IntegrationType, Role, PlatformEntityType, SettingsValueTypes, SettingsTypes,
        SettingsSectionNames,
        SettingScope, BaseSetting, UserSetting, Setting, ServiceSample, ExecutionSample, PipelineExecutionSample,
        ResourceExecution, Message, NotificationEventContext
    )
    from .ml import BaseModelAdapter
    from .utilities import Converter, BaseServiceRunner, Progress, Context, AnnotationFormat
    from .repositories import FUNCTION_END_LINE, PackageCatalog

    def __init__(self, cookie_filepath=None):
        self.client_api = self.ApiClient(cookie_filepath=cookie_filepath)
        self.projects = self.repositories.Projects(client_api=self.client_api)
        self.datasets = self.repositories.Datasets(client_api=self.client_api,
                                                   project=None)
        self.items = self.repositories.Items(client_api=self.client_api,
                                             datasets=self.datasets)
        self.packages = self.repositories.Packages(client_api=self.client_api)
        self.executions = self.repositories.Executions(client_api=self.client_api)
        self.services = self.repositories.Services(client_api=self.client_api)
        self.webhooks = self.repositories.Webhooks(client_api=self.client_api)
        self.triggers = self.repositories.Triggers(client_api=self.client_api)
        self.assignments = self.repositories.Assignments(client_api=self.client_api)
        self.tasks = self.repositories.Tasks(client_api=self.client_api)
        self.dpks = self.repositories.Dpks(client_api=self.client_api)
        self.annotations = self.repositories.Annotations(client_api=self.client_api)
        self.models = self.repositories.Models(client_api=self.client_api)
        self.ontologies = self.repositories.Ontologies(client_api=self.client_api)
        self.recipes = self.repositories.Recipes(client_api=self.client_api)
        self.pipelines = self.repositories.Pipelines(client_api=self.client_api)
        self.pipeline_executions = self.repositories.PipelineExecutions(client_api=self.client_api)
        self.feature_sets = self.repositories.FeatureSets(client_api=self.client_api)
        self.features = self.repositories.Features(client_api=self.client_api)
        self.organizations = self.repositories.Organizations(client_api=self.client_api)
        self.analytics = self.repositories.Analytics(client_api=self.client_api)
        self.integrations = self.repositories.Integrations(client_api=self.client_api)
        self.drivers = self.repositories.Drivers(client_api=self.client_api)
        self.settings = self.repositories.Settings(client_api=self.client_api)
        self.apps = self.repositories.Apps(client_api=self.client_api)
        self.dpks = self.repositories.Dpks(client_api=self.client_api)
        self.messages = self.repositories.Messages(client_api=self.client_api)
        self.compositions = self.repositories.Compositions(client_api=self.client_api)

        self.verbose = self.client_api.verbose
        self.login = self.client_api.login
        self.logout = self.client_api.logout
        self.login_token = self.client_api.login_token
        self.login_secret = self.client_api.login_secret
        self.login_api_key = self.client_api.login_api_key
        self.login_m2m = self.client_api.login_m2m
        self.add_environment = self.client_api.add_environment
        self.setenv = self.client_api.setenv
        self.token_expired = self.client_api.token_expired
        self.info = self.client_api.info
        self.cache_state = self.client_api.cache_state
        self.attributes_mode = self.client_api.attributes_mode
        self.sdk_cache = self.client_api.sdk_cache

    def __del__(self):
        for name, pool in self.client_api._thread_pools.items():
            pool.shutdown()

    def token(self):
        """
        token

        :return: token in use
        """
        return self.client_api.token

    def environment(self):
        """
        environment

        :return: current environment
        """
        return self.client_api.environment

    def init(self):
        """
        init current directory as a Dataloop working directory

        :return:
        """
        from .services import CookieIO
        self.client_api.state_io = CookieIO.init_local_cookie(create=True)
        assert isinstance(self.client_api.state_io, CookieIO)

    def checkout_state(self):
        """
        Return the current checked out state

        :return:
        """
        state = self.client_api.state_io.read_json()
        return state

    class ModalityTypeEnum:
        """
        State enum
        """
        OVERLAY = 'overlay'

    class LinkTypeEnum:
        """
        State enum
        """
        ID = 'id'
        URL = 'url'

    class TriggerResource:
        ITEM = 'Item'
        DATASET = 'Dataset'
        ANNOTATION = 'Annotation'

    class TriggerAction:
        CREATED = 'Created'
        UPDATED = 'Updated'
        DELETED = 'Deleted'

    class TriggerExecutionMode:
        ONCE = 'Once'
        ALWAYS = 'Always'

    class PackageInputType:
        ITEM = 'Item'
        DATASET = 'Dataset'
        ANNOTATION = 'Annotation'
        JSON = 'Json'

    class FiltersResource:
        ITEM = 'items'
        ANNOTATION = 'annotations'

    class FiltersOperations:
        OR = 'or'
        AND = 'and'
        IN = 'in'
        NOT_EQUAL = 'ne'
        EQUAL = 'eq'
        GREATER_THAN = 'gt'
        LESS_THAN = 'lt'

    class FiltersMethod:
        OR = 'or'
        AND = 'and'

    class FiltersOrderByDirection:
        DESCENDING = 'descending'
        ASCENDING = 'ascending'

    class KnownFields:
        DIR = 'dir'
        ANNOTATED = 'annotated'
        FILENAME = 'filename'
        CREATED_AT = 'createdAt'
        UPDATED_AT = 'updatedAt'
        LABEL = 'label'
        NAME = 'name'
        HIDDEN = 'hidden'

    class ExecutionStatus:
        SUCCESS = 'success'
        FAILED = 'failed'
        IN_PROGRESS = 'inProgress'
        CREATED = 'created'

    class HttpMethod:
        GET = 'GET'
        POST = 'POST'
        DELETE = 'DELETE'
        PATCH = 'PATCH'

    class ViewAnnotationOptions:
        JSON = 'json'
        MASK = 'mask'
        INSTANCE = 'instance'

    class AnnotationFormat:
        COCO = 'coco'
        VOC = 'voc'
        YOLO = 'yolo'
        DATALOOP = 'dataloop'

    class InstanceCatalog:
        REGULAR_MICRO = 'regular-micro'
        REGULAR_XS = 'regular-xs'
        REGULAR_S = 'regular-s'
        REGULAR_M = 'regular-m'
        REGULAR_L = 'regular-l'
        HIGHMEM_MICRO = 'highmem-micro'
        HIGHMEM_XS = 'highmem-xs'
        HIGHMEM_S = 'highmem-s'
        HIGHMEM_M = 'highmem-m'
        HIGHMEM_L = 'highmem-l'
        GPU_T4_S = "gpu-t4-s"
        GPU_T4_M = "gpu-t4-m"

    class LoggingLevel:
        DEBUG = 'debug'
        WARNING = 'warning'
        CRITICAL = 'critical'
        INFO = 'info'



import logging
import copy
from .. import exceptions, entities, miscellaneous, _api_reference, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class FeatureSets:
    """
    Feature Sets repository
    """

    URL = '/features/sets'

    def __init__(
        self,
        client_api: ApiClient,
        project_id: str = None,
        project: entities.Project = None,
        model_id: str = None,
        model: entities.Model = None,
        dataset_id: str = None,
        dataset: entities.Dataset = None,
    ):
        self._project = project
        self._project_id = project_id
        self._model = model
        self._model_id = model_id
        self._dataset = dataset
        self._dataset_id = dataset_id
        self._client_api = client_api

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None and self._project_id is None and self.dataset is not None:
            self._project = self.dataset.project
            self._project_id = self._project.id
        if self._project is None and self._project_id is not None:
            # get from id
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)
        if self._project is None:
            # try get checkout
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001', message='Cannot perform action WITHOUT Project entity in FeatureSets repository.' ' Please checkout or set a project'
            )
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def model(self) -> entities.Model:
        if self._model is None and self._model_id is not None:
            # get from id
            self._model = repositories.Models(client_api=self._client_api).get(model_id=self._model_id)
        if self._model is None:
            raise exceptions.PlatformException(error='2001', message='Cannot perform action WITHOUT Model entity in FeatureSets repository.')
        assert isinstance(self._model, entities.Model)
        return self._model

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None and self._dataset_id is not None:
            # get from id
            self._dataset = repositories.Datasets(client_api=self._client_api).get(dataset_id=self._dataset_id)
        if self._dataset is None:
            return None
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    ###########
    # methods #
    ###########

    def _list(self, filters: entities.Filters):
        # request
        success, response = self._client_api.gen_request(req_type='POST', path='/features/sets/query', json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/features/sets/query', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List Feature Sets

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities
        """
        # Step 1: Initialize filter (create empty if None)
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.FEATURE_SET)

        if filters.resource != entities.FiltersResource.FEATURE_SET:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.FEATURE_SET. Got: {!r}'.format(filters.resource),
            )

        # Step 2: Extract IDs inline (no helper functions, no property access)
        # Extract project_id inline
        if self._project_id is not None:
            project_id = self._project_id
        elif self._project is not None:
            project_id = self._project.id
        else:
            raise exceptions.PlatformException(
                error='2001', message='Cannot perform action WITHOUT Project entity in FeatureSets repository.' ' Please checkout or set a project'
            )

        # Extract dataset_id inline (only when needed for filtering)
        if self._dataset_id is not None:
            dataset_id = self._dataset_id
        elif self._dataset is not None:
            dataset_id = self._dataset.id
        else:
            dataset_id = None  # No dataset filtering needed

        # Set context with project_id
        filters.context = {'projects': [project_id]}

        # Preserve original page and page size values before any operations that might modify them
        received_filter = copy.deepcopy(filters)

        # Step 3: Execute received filter - Run filter with appropriate pagination
        # When dataset_id is None: use original pagination (respect user's page request)
        # When dataset_id is not None: start from page 0 to collect all IDs (pagination applied later)
        if dataset_id is not None:
            page_offset = 0
            received_filter_paged = entities.PagedEntities(
                items_repository=self,
                filters=filters,
                page_offset=page_offset,
                page_size=filters.page_size,
                client_api=self._client_api,
            )
            # Step 5: Dataset_id exists and items exist - extract IDs from all pages
            # Extract feature set IDs from all pages (received_filter_paged.all() will fetch all pages starting from page 0)
            filter_fs_ids = [feature_set.id for feature_set in received_filter_paged.all()]
            
            # Step 6: Run aggregation API (when dataset_id exists)
            payload = {
                "projectId": project_id,
                "datasetIds": [dataset_id],
            }
            success, response = self._client_api.gen_request(req_type="POST", path="/features/vectors/project-count-aggregation", json_req=payload)
            if not success:
                raise exceptions.PlatformException(response)
            result = response.json()
            # Extract dataset feature set IDs from response, filtering out entries where count == 0
            dataset_fs_ids = [item['featureSetId'] for item in result if item.get('count', 0) > 0]

            # Step 7: Intersect IDs
            final_fs_ids = list(set(filter_fs_ids).intersection(set(dataset_fs_ids)))

            # Step 8: Final return path - Create filter with intersected IDs (no join needed)
            intersected_ids_filter = entities.Filters(resource=entities.FiltersResource.FEATURE_SET)
            intersected_ids_filter.add(field='id', operator=entities.FiltersOperations.IN, values=final_fs_ids)
            intersected_ids_filter.page = received_filter.page  # Preserve original pagination
            intersected_ids_filter.page_size = received_filter.page_size
            intersected_ids_filter.context = {'projects': [project_id]}

            filter_paged = entities.PagedEntities(
                items_repository=self,
                filters=intersected_ids_filter,
                page_offset=intersected_ids_filter.page,
                page_size=intersected_ids_filter.page_size,
                client_api=self._client_api,
            )
            filter_paged.get_page()

        else:
            page_offset = filters.page if dataset_id is None else 0

            filter_paged = entities.PagedEntities(
                items_repository=self,
                filters=filters,
                page_offset=page_offset,
                page_size=filters.page_size,
                client_api=self._client_api,
            )
            filter_paged.get_page()

        return filter_paged

    @_api_reference.add(path='/features/sets/{id}', method='get')
    def get(self, feature_set_name: str = None, feature_set_id: str = None) -> entities.Feature:
        """
        Get Feature Set object

        :param str feature_set_name: name of the feature set
        :param str feature_set_id: id of the feature set
        :return: Feature object
        """
        if feature_set_id is not None:
            success, response = self._client_api.gen_request(req_type="GET", path="{}/{}".format(self.URL, feature_set_id))
            if not success:
                raise exceptions.PlatformException(response)
            feature_set = entities.FeatureSet.from_json(client_api=self._client_api, _json=response.json())
        elif feature_set_name is not None:
            if not isinstance(feature_set_name, str):
                raise exceptions.PlatformException(error='400', message='feature_set_name must be string')
            filters = entities.Filters(resource=entities.FiltersResource.FEATURE_SET)
            filters.add(field='name', values=feature_set_name)
            feature_sets = self.list(filters=filters)
            if feature_sets.items_count == 0:
                raise exceptions.PlatformException(error='404', message='Feature set not found. name: {!r}'.format(feature_set_name))
            elif feature_sets.items_count > 1:
                # more than one matching project
                raise exceptions.PlatformException(error='404', message='More than one feature_set with same name. Please "get" by id')
            else:
                feature_set = feature_sets.items[0]
        else:
            raise exceptions.PlatformException(error='400', message='Must provide an identifier in inputs, feature_set_name or feature_set_id')
        return feature_set

    @_api_reference.add(path='/features/sets', method='post')
    def create(
        self, name: str, size: int, set_type: str, entity_type: entities.FeatureEntityType, project_id: str = None, model_id: set = None, org_id: str = None
    ):
        """
        Create a new Feature Set

        :param str name: the Feature name
        :param int size: the length of a single vector in the set
        :param str set_type: string of the feature type: 2d, 3d, modelFC, TSNE,PCA,FFT
        :param entity_type: the entity that feature vector is linked to. Use the enum dl.FeatureEntityType
        :param str project_id: the ID of the project where feature set will be created
        :param str model_id: the ID of the model that creates the vectors
        :param str org_id: the ID of the org where feature set will be created
        :return: Feature Set object
        """
        if project_id is None:
            if self._project is None:
                raise ValueError('Must input a project id')
            else:
                project_id = self._project.id

        payload = {'name': name, 'size': size, 'type': set_type, 'project': project_id, 'modelId': model_id, 'entityType': entity_type}
        if org_id is not None:
            payload['org'] = org_id
        success, response = self._client_api.gen_request(req_type="post", json_req=payload, path=self.URL)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.FeatureSet.from_json(client_api=self._client_api, _json=response.json()[0])

    @_api_reference.add(path='/features/sets/{id}', method='delete')
    def delete(self, feature_set_id: str):
        """
        Delete feature vector

        :param str feature_set_id: feature set id to delete
        :return: success
        :rtype: bool
        """

        success, response = self._client_api.gen_request(req_type="delete", path=f"{self.URL}/{feature_set_id}")

        # check response
        if success:
            logger.debug("Feature Set deleted successfully")
            return success
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/features/set/{id}', method='patch')
    def update(self, feature_set: entities.FeatureSet) -> entities.FeatureSet:
        """
         Update a Feature Set

         **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.FeatureSet feature_set: FeatureSet object
        :return: FeatureSet
        :rtype: dtlpy.entities.FeatureSet

        **Example**:

         .. code-block:: python

             dl.feature_sets.update(feature_set='feature_set')
        """
        success, response = self._client_api.gen_request(req_type="patch", path=f"{self.URL}/{feature_set.id}", json_req=feature_set.to_json())
        if not success:
            raise exceptions.PlatformException(response)

        logger.debug("feature_set updated successfully")
        # update dataset labels
        feature_set = entities.FeatureSet.from_json(_json=response.json(), client_api=self._client_api, is_fetched=True)
        return feature_set

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Item]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.FeatureSet._protected_from_json, **{'client_api': self._client_api, '_json': item})
        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items



import logging

from .. import entities, miscellaneous, exceptions, repositories, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Triggers:
    """
    Triggers Repository

    The Triggers class allows users to manage triggers and their properties. Triggers activate services.
    See our documentation for more information on `triggers <https://developers.dataloop.ai/tutorials/faas/concept/chapter/>`_.
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 service: entities.Service = None,
                 project_id: str = None,
                 pipeline: entities.Pipeline = None):
        self._client_api = client_api
        self._project = project
        self._service = service
        self._pipeline = pipeline
        if project_id is None:
            if self._project is not None:
                project_id = self._project.id
            elif self._service is not None:
                project_id = self._service.project_id

        self._project_id = project_id

    ############
    # entities #
    ############
    @property
    def service(self) -> entities.Service:
        if self._service is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "service". need to set a Service entity or use service.triggers repository')
        assert isinstance(self._service, entities.Service)
        return self._service

    @service.setter
    def service(self, service: entities.Service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid Service entity')
        self._service = service

    @property
    def pipeline(self) -> entities.Pipeline:
        if self._pipeline is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "pipeline". need to set a Pipeline entity or use pipeline.triggers repository')
        assert isinstance(self._pipeline, entities.Pipeline)
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline: entities.Pipeline):
        if not isinstance(pipeline, entities.Pipeline):
            raise ValueError('Must input a valid Service entity')
        self._pipeline = pipeline

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._service is not None:
                self._project = self._service._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.triggers repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def name_validation(self, name: str):
        """
        This method validates the trigger name. If name is not valid, this method will return an error. Otherwise, it will not return anything.

        :param str name: trigger name
        """
        url = '/piper-misc/naming/triggers/{}'.format(name)

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/triggers', method='post')
    def create(self,
               # for both trigger types
               service_id: str = None,
               trigger_type: entities.TriggerType = entities.TriggerType.EVENT,
               name: str = None,
               webhook_id=None,
               function_name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
               project_id=None,
               active=True,
               # for event trigger
               filters=None,
               resource: entities.TriggerResource = entities.TriggerResource.ITEM,
               actions: entities.TriggerAction = None,
               execution_mode: entities.TriggerExecutionMode = entities.TriggerExecutionMode.ONCE,
               # for cron triggers
               start_at=None,
               end_at=None,
               inputs=None,
               cron=None,
               pipeline_id=None,
               pipeline=None,
               pipeline_node_id=None,
               root_node_namespace=None,
               **kwargs) -> entities.BaseTrigger:
        """
        Create a Trigger. Can create two types: a cron trigger or an event trigger.
        Inputs are different for each type

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        Inputs for all types:

        :param str service_id: Id of services to be triggered
        :param str trigger_type: can be cron or event. use enum dl.TriggerType for the full list
        :param str name: name of the trigger
        :param str webhook_id: id for webhook to be called
        :param str  function_name: the function name to be called when triggered (must be defined in the package)
        :param str  project_id: project id where trigger will work
        :param bool active: optional - True/False, default = True, if true trigger is active

        Inputs for event trigger:
        :param dtlpy.entities.filters.Filters filters: optional - Item/Annotation metadata filters, default = none
        :param str resource: optional - Dataset/Item/Annotation/ItemStatus, default = Item
        :param str actions: optional - Created/Updated/Deleted, default = create
        :param str execution_mode: how many times trigger should be activated; default is "Once". enum dl.TriggerExecutionMode

        Inputs for cron trigger:
        :param start_at: iso format date string to start activating the cron trigger
        :param end_at: iso format date string to end the cron activation
        :param inputs: dictionary "name":"val" of inputs to the function
        :param str cron: cron spec specifying when it should run. more information: https://en.wikipedia.org/wiki/Cron
        :param str pipeline_id: Id of pipeline to be triggered
        :param pipeline: pipeline entity to be triggered
        :param str pipeline_node_id: Id of pipeline root node to be triggered
        :param root_node_namespace: namespace of pipeline root node to be triggered

        :return: Trigger entity
        :rtype: dtlpy.entities.trigger.Trigger

        **Example**:

        .. code-block:: python

            service.triggers.create(name='triggername',
                                  execution_mode=dl.TriggerExecutionMode.ONCE,
                                  resource='Item',
                                  actions='Created',
                                  function_name='run',
                                  filters={'$and': [{'hidden': False},
                                                    {'type': 'file'}]}
                                  )
        """
        scope = kwargs.get('scope', None)

        if service_id is None and webhook_id is None and pipeline_id is None and pipeline is None:
            if self._service is not None:
                service_id = self._service.id
            elif self._pipeline is not None:
                pipeline = self._pipeline
                pipeline_id = self._pipeline.id

        if pipeline is not None:
            pipeline_id = pipeline.id

        # type
        input_num = sum(input_id is not None for input_id in [service_id, webhook_id, pipeline_id])
        if input_num != 1:
            raise exceptions.PlatformException('400',
                                               'Must provide only one of service id, webhook id, pipeline id or pipeline')

        if pipeline_id is not None:
            if pipeline is None:
                pipeline = repositories.Pipelines(client_api=self._client_api).get(pipeline_id=pipeline_id)
            if pipeline_node_id is None:
                if pipeline.start_nodes:
                    for pipe_node in pipeline.start_nodes:
                        if pipe_node['type'] == 'root':
                            pipeline_node_id = pipe_node['nodeId']
            if pipeline_node_id is None:
                raise exceptions.PlatformException('400', 'Must provide pipeline node id')
            if not actions:
                actions = [entities.TriggerAction.CREATED]
            pipeline.nodes.get(node_id=pipeline_node_id).add_trigger(
                trigger_type=trigger_type,
                filters=filters,
                resource=resource,
                actions=actions,
                execution_mode=execution_mode,
                cron=cron,
            )
            logger.info("The trigger will not create until pipeline is install")
            pipeline.update()
            return True
        else:
            if name is None:
                if self._service is not None:
                    name = self._service.name
                else:
                    name = 'defaulttrigger'

            if filters is None:
                filters = dict()
            elif isinstance(filters, entities.Filters):
                filters = filters.prepare(query_only=True).get('filter', dict())

            if webhook_id is not None:
                operation = {
                    'type': 'webhook',
                    'webhookId': webhook_id
                }
            else:
                operation = {
                    'type': 'function',
                    'serviceId': service_id,
                    'functionName': function_name

                }

            if actions is not None:
                if not isinstance(actions, list):
                    actions = [actions]
            else:
                actions = [entities.TriggerAction.CREATED]

            if len(actions) == 0:
                actions = [entities.TriggerAction.CREATED]

            if trigger_type == entities.TriggerType.EVENT:
                spec = {
                    'filter': filters,
                    'resource': resource,
                    'executionMode': execution_mode,
                    'actions': actions
                }
            elif trigger_type == entities.TriggerType.CRON:
                spec = {
                    'endAt': end_at,
                    'startAt': start_at,
                    'cron': cron,
                }
            else:
                raise ValueError('Unknown trigger type: "{}". Use dl.TriggerType for known types'.format(trigger_type))

            spec['input'] = dict() if inputs is None else inputs
            spec['operation'] = operation

            # payload
            if self._project_id is None and project_id is None:
                raise exceptions.PlatformException('400', 'Please provide a project id')
            elif project_id is None:
                project_id = self._project_id

            payload = {
                'type': trigger_type,
                'active': active,
                'projectId': project_id,
                'name': name,
                'spec': spec
            }

            if scope is not None:
                logger.warning(
                    "Only superuser is allowed to define a trigger's scope. "
                    "If you are not a superuser you will not be able to perform this action")
                payload['scope'] = scope

            # request
            success, response = self._client_api.gen_request(req_type='post',
                                                             path='/triggers',
                                                             json_req=payload)

            # exception handling
            if not success:
                raise exceptions.PlatformException(response)

            # return entity
            return entities.BaseTrigger.from_json(_json=response.json(),
                                                  client_api=self._client_api,
                                                  project=self._project if self._project_id == project_id else None,
                                                  service=self._service)

    @_api_reference.add(path='/triggers/{id}', method='get')
    def get(self, trigger_id=None, trigger_name=None) -> entities.BaseTrigger:
        """
        Get Trigger object

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str trigger_id: trigger id
        :param str  trigger_name: trigger name
        :return: Trigger entity
        :rtype: dtlpy.entities.trigger.Trigger

        **Example**:

        .. code-block:: python

            service.triggers.get(trigger_id='trigger_id')
        """
        # request
        if trigger_id is not None:
            success, response = self._client_api.gen_request(
                req_type="get",
                path="/triggers/{}".format(trigger_id)
            )

            # exception handling
            if not success:
                raise exceptions.PlatformException(response)

            # return entity
            trigger = entities.BaseTrigger.from_json(client_api=self._client_api,
                                                     _json=response.json(),
                                                     project=self._project,
                                                     service=self._service)
            # verify input trigger name is same as the given id
            if trigger_name is not None and trigger.name != trigger_name:
                logger.warning(
                    "Mismatch found in triggers.get: trigger_name is different then trigger.name:"
                    " {!r} != {!r}".format(
                        trigger_name,
                        trigger.name))
        else:
            if trigger_name is None:
                raise exceptions.PlatformException('400', 'Must provide either trigger name or trigger id')
            else:
                filters = self.__generate_default_filter()
                filters.add(field='name', values=trigger_name)
                triggers = self.list(filters)
                if triggers.items_count == 0:
                    raise exceptions.PlatformException('404', 'Trigger not found')
                elif triggers.items_count == 1:
                    trigger = triggers.items[0]
                else:
                    raise exceptions.PlatformException('404',
                                                       'More than one trigger by name {} exist'.format(trigger_name))

        return trigger

    @_api_reference.add(path='/triggers/{id}', method='delete')
    def delete(self, trigger_id=None, trigger_name=None):
        """
        Delete Trigger object

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str trigger_id: trigger id
        :param str trigger_name: trigger name
        :return: True is successful error if not
        :rtype: bool

        **Example**:

        .. code-block:: python

            service.triggers.delete(trigger_id='trigger_id')
        """
        if trigger_id is None:
            if trigger_name is None:
                raise exceptions.PlatformException('400', 'Must provide either trigger name or trigger id')
            else:
                trigger_id = self.get(trigger_name=trigger_name).id
        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/triggers/{}".format(trigger_id)
        )
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        return True

    @_api_reference.add(path='/triggers/{id}', method='patch')
    def update(self, trigger: entities.BaseTrigger) -> entities.BaseTrigger:
        """
        Update trigger

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.trigger.Trigger trigger: Trigger entity
        :return: Trigger entity
        :rtype: dtlpy.entities.trigger.Trigger

        **Example**:

        .. code-block:: python

            service.triggers.update(trigger='trigger_entity')
        """
        # payload
        payload = trigger.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/triggers/{}'.format(trigger.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.BaseTrigger.from_json(_json=response.json(),
                                              client_api=self._client_api,
                                              project=self._project,
                                              service=self._service)

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.BaseTrigger]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_trigger, trigger in enumerate(response_items):
            jobs[i_trigger] = pool.submit(entities.BaseTrigger._protected_from_json,
                                          **{'client_api': self._client_api,
                                             '_json': trigger,
                                             'project': self._project,
                                             'service': self._service})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        triggers = miscellaneous.List([r[1] for r in results if r[0] is True])
        return triggers

    def _list(self, filters: entities.Filters):
        """
        List project triggers
        :return:
        """
        url = '/query/faas'

        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def __generate_default_filter(self):
        filters = entities.Filters(resource=entities.FiltersResource.TRIGGER)
        if self._project is not None:
            filters.add(field='projectId', values=self._project.id)
        if self._service is not None:
            filters.add(field='spec.operation.serviceId', values=self._service.id)
        if self._pipeline is not None:
            filters.add(field='spec.operation.id', values=self._pipeline.id)

        return filters

    @_api_reference.add(path='/query/faas', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List triggers of a project, package, or service.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            service.triggers.list()
        """
        if filters is None:
            filters = self.__generate_default_filter()
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))

        if filters.resource != entities.FiltersResource.TRIGGER:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.TRIGGER. Got: {!r}'.format(filters.resource))

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def resource_information(self, resource, resource_type, action='Created'):
        """
        Returns which function should run on an item (based on global triggers).

        **Prerequisites**: You must be a **superuser** to run this method.

        :param resource: 'Item' / 'Dataset' / etc
        :param resource_type: dictionary of the resource object
        :param action: 'Created' / 'Updated' / etc.

        **Example**:

        .. code-block:: python

            service.triggers.resource_information(resource='Item', resource_type=item_object, action='Created')
        """
        url = '/trigger-resource-information'

        payload = {'resource': resource_type,
                   'entity': resource.to_json(),
                   'action': action}
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()



import logging
import math
import time
import tqdm
import copy
import sys
from typing import Optional, List, Any

import attr

from .filters import FiltersOperations, FiltersOrderByDirection, FiltersResource
from .. import miscellaneous, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


@attr.s
class PagedEntities:
    """
    Pages object for efficient API pagination.
    Defaults to offset-based pagination for compatibility with all operations.
    Switches to keyset/cursor-based pagination (using 'id' as the cursor) during iteration for performance.
    Falls back to offset-based pagination if keyset is not possible (e.g., custom sort).
    """
    # api
    _client_api: ApiClient = attr.ib(repr=False)

    # params
    page_offset: int = attr.ib()
    page_size: int = attr.ib()
    filters: Any = attr.ib()
    items_repository: Any = attr.ib(repr=False)
    has_next_page: bool = attr.ib(default=False)
    total_pages_count: int = attr.ib(default=0)
    items_count: int = attr.ib(default=0)

    # hybrid pagination
    use_id_based_paging: bool = attr.ib(default=False)  # Default to False for offset-based pagination
    last_seen_id: Optional[Any] = attr.ib(default=None)

    # execution attribute
    _service_id = attr.ib(default=None, repr=False)
    _project_id = attr.ib(default=None, repr=False)
    _list_function = attr.ib(default=None, repr=False)

    # items list
    items: List[Any] = attr.ib(default=miscellaneous.List(), repr=False)

    @staticmethod
    def _has_explicit_sort(flt) -> bool:
        """
        Check if the filter has custom sort fields defined (not id/createdAt).
        """
        prepared = flt.prepare() if flt else {}
        sort_fields = list(prepared.get("sort", {}).keys())
        if isinstance(sort_fields, list) and len(sort_fields) > 0:
            return sort_fields[0] not in {"id", "createdAt"}
        return False

    def _should_use_keyset_pagination(self) -> bool:
        """
        Determine whether to use keyset pagination based on page offset and resource type.
        Keyset pagination can only be used when page_offset is 0 (first page).
        :param page_offset: The page offset to check
        :return: True if keyset pagination should be used, False otherwise
        """
        # Keyset pagination only works for page 0 (first page)
        if self.page_offset != 0:
            return False
                
        # can't use add to custom filter
        if self.filters.custom_filter is not None:
            return False
        
        # Check if the resource supports keyset pagination
        enable_id_based_paging = getattr(self.filters, "resource", None) in [
            FiltersResource.ITEM,
            FiltersResource.ANNOTATION,
            FiltersResource.FEATURE,
        ]
        
        if not enable_id_based_paging:
            return False
            
        # Check if there's no explicit sort that would prevent keyset pagination
        if self._has_explicit_sort(self.filters):
            return False
            
        return True

    def process_result(self, result: dict) -> List[Any]:
        """
        Process the API result and update pagination state.
        :param result: json object
        :return: list of items
        """
        # Only update page_offset if using offset-based pagination
        if not self.use_id_based_paging and 'page_offset' in result:
            self.page_offset = result['page_offset']
        if 'page_size' in result:
            self.page_size = result['page_size']
        if 'hasNextPage' in result:
            self.has_next_page = result['hasNextPage']
        if 'totalItemsCount' in result:
            self.items_count = result['totalItemsCount']
        if 'totalPagesCount' in result:
            self.total_pages_count = result['totalPagesCount']
        if 'items' in result:
            items = self.items_repository._build_entities_from_response(response_items=result['items'])
        else:
            items = miscellaneous.List(list())
        return items

    def __getitem__(self, y: int) -> List[Any]:
        # If we're already on the requested page, return current items
        if y == self.page_offset:
            return self.items
        # Otherwise, go to the requested page
        self.go_to_page(y)
        return self.items

    def __len__(self) -> int:
        return self.items_count

    def __iter__(self):
        # Use keyset/cursor-based pagination for iteration when possible
        self.last_seen_id = None
        self.page_offset = 0  # Start from the first page for iteration
        self.use_id_based_paging = self._should_use_keyset_pagination()
        self.has_next_page = True  # Start with assumption that there are more pages
        self.page_size = self.page_size or 100
        pbar = tqdm.tqdm(total=self.items_count,
                         disable=self._client_api.verbose.disable_progress_bar_iterate_pages,
                         file=sys.stdout, desc="Iterate Pages")
        
        # Get the first page
        self.get_page()
        if self.items:
            yield self.items
            pbar.update()
        
        # Continue with next pages
        while self.has_next_page:
            if self.use_id_based_paging:
                # For keyset pagination, just get the next page
                self.page_offset = 0
                self.get_page()
            else:
                # For offset pagination, increment the offset
                self._move_page_offset(1)
                self.get_page()
            
            if not self.items:
                break
            yield self.items
            pbar.update()
        pbar.close()

    def __reversed__(self):
        # Force offset-based pagination for reverse iteration
        self.use_id_based_paging = False
        self.page_offset = self.total_pages_count - 1
        while True:
            self.get_page()
            yield self.items
            if self.page_offset == 0:
                break
            self._move_page_offset(-1)

    def _move_page_offset(self, offset: int) -> None:
        """
        Move the page offset by a given step.
        :param offset: offset to move
        """
        self.page_offset += offset
        if self.filters.custom_filter is not None:
            if 'page' in self.filters.custom_filter and self.filters.custom_filter['page'] != self.page_offset:
                self.filters.custom_filter['page'] = self.page_offset

    def return_page(self, page_offset: Optional[int] = None, page_size: Optional[int] = None) -> List[Any]:
        """
        Return a page of results using offset-based pagination by default.
        Switches to keyset/cursor-based pagination when supported and beneficial.
        :param page_offset: page offset (for offset-based)
        :param page_size: page size
        :return: list of items
        """
        if page_size is not None:
            self.page_size = page_size
        if page_offset is not None:
            self.page_offset = page_offset

        if self.filters is None:
            raise ValueError("Can't return page. Filters is empty")
        self.filters.page_size = self.page_size
        self.filters.page = self.page_offset
        req = copy.deepcopy(self.filters)

        # Determine pagination method based on page offset and resource type
        self.use_id_based_paging = self._should_use_keyset_pagination()

        if self.use_id_based_paging:
            # Use keyset/cursor-based pagination
            prepared = req.prepare()
            sort_spec = prepared.get("sort", {})
            order = next(iter(sort_spec.values()), None)
            if order is None:
                order = FiltersOrderByDirection.ASCENDING
            if order == FiltersOrderByDirection.DESCENDING:
                operator_value = FiltersOperations.LESS_THAN
            else:
                operator_value = FiltersOperations.GREATER_THAN
                
            req.sort_by(field="id", value=order)
            req.page = 0  # always fetch from the start for keyset
            # Only add last_seen_id filter if we're not explicitly requesting page 0
            if self.last_seen_id:
                req.add(
                    field="id",
                    values=self.last_seen_id,
                    operator=operator_value,
                    method=FiltersOperations.AND,
                )
        # Fetch data
        if self._list_function is None:
            result = self.items_repository._list(filters=req)
        else:
            result = self._list_function(filters=req)

        items = self.process_result(result)

        # Update last_seen_id for keyset
        if self.use_id_based_paging and items and hasattr(items[-1], "id"):
            self.last_seen_id = items[-1].id
        elif self.use_id_based_paging and not items:
            self.last_seen_id = None
        return items

    def get_page(self, page_offset: Optional[int] = None, page_size: Optional[int] = None) -> None:
        """
        Get a page of results and update self.items.
        :param page_offset: page offset (for offset-based)
        :param page_size: page size
        """
        try:
            items = self.return_page(page_offset=page_offset, page_size=page_size)
            self.items = items
        except exceptions.BadRequest as e:
            logger.warning(f"BadRequest error received: {str(e)}")
            self.items = miscellaneous.List(list())

    def next_page(self) -> None:
        """
        Brings the next page of items from host.
        """
        if self.use_id_based_paging:
            # For keyset pagination, just get the next page
            self.get_page()
        else:
            # For offset pagination, increment the offset
            self._move_page_offset(1)
            self.get_page()

    def prev_page(self) -> None:
        """
        Brings the previous page of items from host.
        Only works with offset-based pagination.
        """
        if self.use_id_based_paging:
            raise NotImplementedError("prev_page is not supported for keyset pagination.")
        self._move_page_offset(-1)
        self.get_page()

    def go_to_page(self, page: int = 0) -> None:
        """
        Brings specified page of items from host.
        For page 0, uses keyset pagination if supported.
        For other pages, uses offset-based pagination.
        :param page: page number
        """
        # Reset last_seen_id when going to page 0 to ensure we get all items
        if page == 0:
            self.last_seen_id = None
        self.page_offset = page
        self.get_page()

    def all(self):
        """
        Iterate over all items in all pages efficiently.
        Uses the iterator implementation (__iter__).
        """
        for items in self:
            for item in items:
                yield item

    ########
    # misc #
    ########
    def print(self, columns=None):
        self.items.print(columns=columns)

    def to_df(self, columns=None):
        return self.items.to_df(columns=columns)